from spydrnet.ir import Element, FirstClassElement, InnerPin, OuterPin, Wire, Netlist, Library, Definition, Instance,\
    Port, Cable
from spydrnet.util.hierarchical_reference import HRef
from spydrnet.global_state.global_service import lookup
from spydrnet.util.patterns import _is_pattern_absolute, _value_matches_pattern


def get_ports(obj, *args, **kwargs):
    """
    get_ports(obj, ...)

    Get ports *within* an object.

    Parameters
    ----------
    obj : object, Iterable - required
        The object or objects associated with this query. Queries return a collection objects associated with the
        provided object or objects that match the query criteria. For example, `sdn.get_ports(definition, ...)` would
        return all of the ports associated with the provided definition that match the additional criteria.
    patterns : str, Iterable - optional, positional or named, default: wildcard
        The search patterns. Patterns can be a single string or an Iterable collection of strings. Patterns can be
        absolute or they can contain wildcards or regular expressions. If `patterns` is not provided, then it defaults
        to a wildcard. Patterns are queried against the object property value stored under a specified key. Fast lookups
        are only attempted on absolute patterns that are not regular expressions and contain no wildcards.
    key : str, optional, default: ".NAME"
        This is the key that controls which value is being searched.
    is_case : bool - optional, named, default: True
        Specify if patterns should be treated as case sensitive. Only applies to patterns. Does not alter fast lookup
        behavior (if namespace policy uses case insensitive indexing, this parameter will not prevent a fast lookup
        from returning a matching object even if the case is not an exact match).
    is_re: bool - optional, named, default: False
        Specify if patterns are regular expressions. If `False`, a pattern can still contain `*` and `?` wildcards. A
        `*` matches zero or more characters. A `?` matches upto a single character.
    filter : function
        This is a single input function that can be used to filter out unwanted virtual instances. If not specifed, all
        matching virtual instances are returned. Otherwise, virtual instances that cause the filter function to evaluate
        to true are the only items returned.
    
    Returns
    -------
    ports : generator
        The ports associated with a particular object or collection of objects.
    
    """
    # Check argument list
    if len(args) == 1 and 'patterns' in kwargs:
        raise TypeError("get_ports() got multiple values for argument 'patterns'")
    if len(args) > 1 or any(x not in {'patterns', 'key', 'filter', 'is_case', 'is_re'} for x in kwargs):
        raise TypeError("Unknown usage. Please see help for more information.")

    # Default values
    filter_func = kwargs.get('filter', lambda x: True)
    is_case = kwargs.get('is_case', True)
    is_re = kwargs.get('is_re', False)
    patterns = args[0] if len(args) == 1 else kwargs.get('patterns', ".*" if is_re else "*")
    key = kwargs.get('key', ".NAME")

    if isinstance(obj, (FirstClassElement, InnerPin, OuterPin, Wire)) is False:
        try:
            object_collection = list(iter(obj))
        except TypeError:
            object_collection = [obj]
    else:
        object_collection = [obj]
    if all(isinstance(x, (Element, HRef)) for x in object_collection) is False:
        raise TypeError("get_ports() supports netlist elements and hierarchical references, or a collection of "
                        "these as the object searched")

    if isinstance(patterns, str):
        patterns = (patterns,)

    return _get_ports(object_collection, patterns, key, is_case, is_re, filter_func)


def _get_ports(object_collection, patterns, key, is_case, is_re, filter_func):
    for result in filter(filter_func, _get_ports_raw(object_collection, patterns, key, is_case, is_re)):
        yield result


def _get_ports_raw(object_collection, patterns, key, is_case, is_re):
    found = set()
    other_ports = set()
    while object_collection:
        obj = object_collection.pop()
        if isinstance(obj, Definition):
            for pattern in patterns:
                pattern_is_absolute = _is_pattern_absolute(pattern, is_case, is_re)
                if pattern_is_absolute:
                    result = lookup(obj, Port, key, pattern)
                    if result is not None and result not in found:
                        found.add(result)
                        yield result
                else:
                    for port in obj.ports:
                        value = port[key] if key in port else ''
                        if port not in found and _value_matches_pattern(value, pattern, is_case, is_re):
                            found.add(port)
                            yield port
        elif isinstance(obj, Netlist):
            for library in obj.libraries:
                for definition in library.definitions:
                    object_collection.append(definition)
        elif isinstance(obj, Library):
            object_collection += obj.definitions
        elif isinstance(obj, Instance):
            reference = obj.reference
            if reference:
                object_collection.append(reference)
        elif isinstance(obj, Port):
            other_ports.add(obj)
        elif isinstance(obj, InnerPin):
            port = obj.port
            if port:
                other_ports.add(port)
        elif isinstance(obj, OuterPin):
            inner_pin = obj.inner_pin
            if inner_pin:
                object_collection.append(inner_pin)
        elif isinstance(obj, Wire):
            object_collection += obj.pins
        elif isinstance(obj, Cable):
            object_collection += obj.wires
        elif isinstance(obj, HRef):
            if obj.is_valid:
                object_collection.append(obj.item)

    if other_ports:
        namemap = dict()
        for other_port in other_ports:
            if other_port in found:
                continue
            found.add(other_port)
            name = other_port[key] if key in other_port else ''
            if name not in namemap:
                namemap[name] = list()
            namemap[name].append(other_port)
        for pattern in patterns:
            pattern_is_absolute = _is_pattern_absolute(pattern, is_case, is_re)
            if pattern_is_absolute:
                if pattern in namemap:
                    result = namemap[pattern]
                    del namemap[pattern]
                    for port in result:
                        yield port
            else:
                names_to_remove = list()
                for name in namemap:
                    if _value_matches_pattern(name, pattern, is_case, is_re):
                        result = namemap[name]
                        names_to_remove.append(name)
                        for port in result:
                            yield port
                for name in names_to_remove:
                    del namemap[name]



import spydrnet as sdn


TAB = '\t'
NEW_LINE ='\n'
CARRIAGE_RETURN = '\r'
SPACE = ' '
FORM_FEED = '\f'

NUMBERS = {'0','1','2','3','4','5','6','7','8','9'}
LETTERS = {'a','b','c','d','e','f','g','h','i','j',\
    'k','l','m','n','o','p','q','r','s','t','u','v',\
    'w','x','y','z',\
    'A','B','C','D','E','F','G','H',\
    'I','J','K','L','M','N','O','P','Q','R','S','T',\
    'U','V','W','X','Y','Z'}

WHITESPACE = {SPACE, TAB, NEW_LINE, CARRIAGE_RETURN, FORM_FEED}

OPEN_PARENTHESIS = '('
CLOSE_PARENTHESIS = ')'
STAR = '*'
SEMI_COLON = ';'
MODULE = "module"
INPUT = "input"
OUTPUT = "output"
INOUT = "inout"
WIRE = "wire"
REG = "reg"
DOT = "."
OPEN_BRACKET = "["
CLOSE_BRACKET = "]"
OPEN_BRACE = "{"
CLOSE_BRACE = "}"
COLON = ":"
COMMA = ","
OCTOTHORP = "#"
END_MODULE = "endmodule"
PARAMETER = "parameter"
SINGLE_QUOTE = "'"
LOCAL_PARAM = "localparam"
ASSIGN = "assign"
EQUAL = "="
CELL_DEFINE = "`celldefine"
END_CELL_DEFINE = "`endcelldefine"
IFDEF = "`ifdef"
DEFINE = "`define"
ENDIF = "`endif"
ELSIF = '`elsif'
TIMESCALE = '`timescale'
OPEN_BLOCK_COMMENT = "/*"
CLOSE_BLOCK_COMMENT = "*/"
OPEN_LINE_COMMENT = "//"
PRIMITIVE = 'primitive'
END_PRIMITIVE = 'endprimitive'
FUNCTION = 'function'
END_FUNCTION = 'endfunction'
TASK = 'task'
END_TASK = 'endtask'
INTEGER = 'integer'
TRI0 = 'tri0'
TRI1 = 'tri1'
DEFPARAM = 'defparam'

#SET OF ALL THINGS THAT WILL END AN IDENTIFIER IF THEY ARE NOT ESCAPED.
#elif ch in {'(', ')', '.', ',', ';', '[', ']', ':', "{", "}", "*", "#", "`"}:
BREAKER_TOKENS = {SPACE, TAB, NEW_LINE, CARRIAGE_RETURN, FORM_FEED, OPEN_PARENTHESIS, CLOSE_PARENTHESIS,\
    COMMA, SEMI_COLON, OPEN_BRACKET, CLOSE_BRACKET, COLON, OPEN_BRACE, CLOSE_BRACE, STAR, OCTOTHORP,\
        EQUAL, "\\", "\"", "`"} #single quote should not be included here because of 1'b0 type of declarations (these should be one token)

SINGLE_CHARACTER_TOKENS = {OPEN_PARENTHESIS, CLOSE_PARENTHESIS, STAR, SEMI_COLON, DOT, OPEN_BRACKET,\
    CLOSE_BRACKET, OPEN_BRACE, CLOSE_BRACE, COLON, COMMA, COMMA, OCTOTHORP, SINGLE_QUOTE, EQUAL}

PORT_DIRECTIONS = {INPUT, OUTPUT, INOUT}

ALL_VERILOG_TOKENS = {
    TAB,\
    NEW_LINE,\
    CARRIAGE_RETURN,\
    SPACE,\
    FORM_FEED,\
    OPEN_PARENTHESIS,\
    CLOSE_PARENTHESIS,\
    STAR,\
    SEMI_COLON,\
    MODULE,\
    INPUT,\
    OUTPUT,\
    INOUT,\
    WIRE,\
    REG,\
    DOT,\
    OPEN_BRACKET,\
    CLOSE_BRACKET,\
    OPEN_BRACE,\
    CLOSE_BRACE,\
    COLON,\
    COMMA,\
    OCTOTHORP,\
    END_MODULE,\
    PARAMETER,\
    SINGLE_QUOTE,\
    LOCAL_PARAM,\
    ASSIGN,\
    EQUAL,\
    CELL_DEFINE,\
    END_CELL_DEFINE,\
    IFDEF,\
    ENDIF,\
    ELSIF,\
    OPEN_BLOCK_COMMENT,\
    CLOSE_BLOCK_COMMENT,\
    OPEN_LINE_COMMENT,\
    INTEGER,\
    TRI0,\
    TRI1,\
    DEFPARAM
}

def is_valid_identifier(token):
    if token == "":
        return False
    elif token[0] == '\\' and token[-1] == " ":
        for white in WHITESPACE: #there shouldn't be whitespace before the ending space
            if white in token[:-1]:
                return False
        return True
    else:
        if token[0] not in LETTERS and token[0] != "_":
            return False
        for c in token:
            if c not in LETTERS and c not in NUMBERS and c != "_":
                return False
        return True

def is_numeric(token):
    if token == "":
        return False
    for c in token:
        if c not in NUMBERS:
            return False
    return True

def string_to_port_direction(token):
    if token == INPUT:
        return sdn.Port.Direction.IN
    elif token == OUTPUT:
        return sdn.Port.Direction.OUT
    elif token == INOUT:
        return sdn.Port.Direction.INOUT
    else:
        return sdn.Port.Direction.UNDEFINED




import unittest
import spydrnet as sdn


class TestNamespaceManager(unittest.TestCase):

    def test_naming_after_assignments(self):
        library = sdn.Library()
        definition = library.create_definition()
        definition2 = library.create_definition()
        definition.name = "Hello"
        definition.name = "World"
        definition2.name = "Hello"
        caught_conflict = False
        try:
            definition2.name = "World"
        except ValueError:
            caught_conflict = True
        self.assertTrue(caught_conflict)



from abc import ABC

from spydrnet.callback.callback_listener import CallbackListener
from spydrnet.plugins.namespace_manager.default_namespace import DefaultNamespace
from spydrnet.plugins.namespace_manager.edif_namespace import EdifNamespace
from spydrnet.ir import Netlist, Library, Definition, Port, Cable, Instance
from spydrnet.global_state.global_service import register_lookup, deregister_lookup
import weakref


def _load_policies():
    NamespaceManager.policies["DEFAULT"] = DefaultNamespace
    NamespaceManager.policies["EDIF"] = EdifNamespace
    NamespaceManager.default = "DEFAULT"


class NamespaceManager(CallbackListener, ABC):
    default = None
    policies = dict()

    def __init__(self):
        super().__init__()
        self.namespaces = weakref.WeakKeyDictionary()
        self.ignore_ns_change = False

    def register_all_listeners(self):
        super().register_all_listeners()
        register_lookup(".NAME", self.lookup)
        register_lookup("EDIF.identifier", self.lookup)

    def deregister_all_listeners(self):
        super().deregister_all_listeners()
        deregister_lookup(".NAME")
        deregister_lookup("EDIF.identifier")

    def lookup(self, parent, element_type, key, value):
        if parent in self.namespaces:
            namespace = self.namespaces[parent]
            return namespace.lookup(element_type, key, value)

    def create_netlist(self, netlist):
        netlist[".NS"] = self.default

    def create_library(self, library):
        library[".NS"] = self.default

    def create_definition(self, definition):
        definition[".NS"] = self.default

    def create_port(self, port):
        port[".NS"] = self.default

    def create_cable(self, cable):
        cable[".NS"] = self.default

    def create_instance(self, instance):
        instance[".NS"] = self.default

    def definition_add_port(self, definition, port):
        self.add(definition, port)

    def definition_remove_port(self, definition, port):
        self.remove(port, parent=definition)

    def definition_add_child(self, definition, child):
        self.add(definition, child)

    def definition_remove_child(self, definition, child):
        self.remove(child, parent=definition)

    def definition_add_cable(self, definition, cable):
        self.add(definition, cable)

    def definition_remove_cable(self, definition, cable):
        self.remove(cable, parent=definition)

    def library_add_definition(self, library, definition):
        self.add(library, definition)

    def library_remove_definition(self, library, definition):
        self.remove(definition, parent=library)

    def netlist_add_library(self, netlist, library):
        self.add(netlist, library)

    def netlist_remove_library(self, netlist, library):
        self.remove(library, parent=netlist)

    def dictionary_set(self, element, key, value):
        if key == ".NS":
            if self.ignore_ns_change is False and (key not in element or element[key] != value):
                if self.get_parent(element) is not None:
                    raise ValueError("Cannot change the namespace of a object already belonging to a parent")
                if value not in self.policies:
                    raise ValueError('The namespace policy specified does not exist. Supported namespaces include: {}'.format(", ".join(self.policies)))
                target_namespace = self.policies[value]
                if self.is_compliant(target_namespace, element) is False:
                    raise ValueError("The current element is not compliant with the target namespace policy.")
                self.apply_namespace(value, target_namespace, element)
        elif key in {".NAME", "EDIF.identifier"}:
            if ".NS" in element:
                target_policy = self.policies[element[".NS"]]
                if target_policy.is_name_valid(key, value) is False:
                    raise ValueError("Target name not valid for the current namespace policy.")

            parent = self.get_parent(element)
            if parent and parent in self.namespaces:
                namespace = self.namespaces[parent]
                if namespace.no_conflict(element, key, value):
                    namespace.update(element, key, value)
                else:
                    raise ValueError("Applying name would result in a naming conflict")

    def dictionary_delete(self, element, key):
        if key == ".NS":
            if self.ignore_ns_change is False:
                if self.get_parent(element) is not None:
                    raise ValueError("Cannot change the namespace of a object already belonging to a parent")
                if ".NS" in element:
                    self.drop_namespace(element)
        elif key in {".NAME", "EDIF.identifier"}:
            parent = self.get_parent(element)
            if parent and parent in self.namespaces:
                namespace = self.namespaces[parent]
                namespace.remove(element, key)

    def dictionary_pop(self, element, key):
        if key == ".NS":
            if self.ignore_ns_change is False:
                if self.get_parent(element) is not None:
                    raise ValueError("Cannot change the namespace of a object already belonging to a parent")
                if ".NS" in element:
                    self.drop_namespace(element)
        elif key in {".NAME", "EDIF.identifier"}:
            self.remove(element, key)

    def add(self, parent, child):
        if parent:
            namespace = None
            if parent in self.namespaces:
                namespace = self.namespaces[parent]
                for key in ["EDIF.identifier", ".NAME"]:
                    if key in child:
                        no_conflict = namespace.no_conflict(child, key, child[key])
                        if no_conflict is False:
                            raise ValueError("Adding this element would result in a naming conflict. " + child[key])
            if ".NS" in parent:
                parent_namespace = parent[".NS"]
                if ".NS" not in child or child[".NS"] != parent_namespace:
                    child[".NS"] = parent_namespace
            elif ".NS" in child:
                del child[".NS"]
            if namespace is not None:
                for key in ["EDIF.identifier", ".NAME"]:
                    if key in child:
                        namespace.update(child, key, child[key])

    def remove(self, element, key=None, parent=None):
        if parent is None:
            parent = self.get_parent(element)
        if parent and parent in self.namespaces:
            namespace = self.namespaces[parent]
            if key is None:
                for key in ["EDIF.identifier", ".NAME"]:
                    if key in element:
                        namespace.remove(element, key)
            else:
                namespace.remove(element, key)

    @staticmethod
    def get_parent(element):
        parent = None
        if isinstance(element, Library):
            parent = element.netlist
        elif isinstance(element, Definition):
            parent = element.library
        elif isinstance(element, (Port, Cable)):
            parent = element.definition
        elif isinstance(element, Instance):
            parent = element.parent
        return parent

    @staticmethod
    def is_compliant(target_namespace, element):
        search_stack = [(element, False)]
        while search_stack:
            element, visited = search_stack.pop()
            if not visited:
                search_stack.append((element, True))
                if isinstance(element, Netlist):
                    search_stack += ((x, False) for x in element.libraries)
                elif isinstance(element, Library):
                    search_stack += ((x, False) for x in element.definitions)
                elif isinstance(element, Definition):
                    search_stack += ((x, True) for x in element.ports)
                    search_stack += ((x, True) for x in element.cables)
                    search_stack += ((x, True) for x in element.children)
            else:
                if target_namespace.is_compliant(element) is False:
                    return False
        return True

    def apply_namespace(self, value, target_namespace, original_element):
        self.ignore_ns_change = True
        search_stack = [original_element]
        while search_stack:
            element = search_stack.pop()
            if ".NS" in element and element in self.namespaces:
                del self.namespaces[element]
            element[".NS"] = value
            new_namespace = None
            if target_namespace.needs_namespace(element):
                new_namespace = target_namespace()
                self.namespaces[element] = new_namespace

            if isinstance(element, Netlist):
                self._update_new_namespace(element.libraries, new_namespace)
                search_stack += element.libraries
            elif isinstance(element, Library):
                self._update_new_namespace(element.definitions, new_namespace)
                search_stack += element.definitions
            elif isinstance(element, Definition):
                self._update_new_namespace(element.ports, new_namespace)
                self._update_new_namespace(element.cables, new_namespace)
                self._update_new_namespace(element.children, new_namespace)
                search_stack += element.ports
                search_stack += element.cables
                search_stack += element.children
        self.ignore_ns_change = False

    def _update_new_namespace(self, elements, namespace):
        if namespace:
            for element in elements:
                if element.name:
                    namespace.update(element, ".NAME", element.name)
                if 'EDIF.identifier' in element:
                    namespace.update(element, "EDIF.identifier", element['EDIF.identifier'])

    def drop_namespace(self, original_element):
        self.ignore_ns_change = True
        search_stack = [original_element]
        while search_stack:
            element = search_stack.pop()
            if element in self.namespaces:
                del self.namespaces[element]
            if element is not original_element and ".NS" in element:
                del element[".NS"]
            if isinstance(element, Netlist):
                search_stack += element.libraries
            elif isinstance(element, Library):
                search_stack += element.definitions
            elif isinstance(element, Definition):
                search_stack += element.ports
                search_stack += element.cables
                search_stack += element.children
        self.ignore_ns_change = False


_load_policies()



import unittest
import spydrnet as sdn
from spydrnet import composers
from spydrnet import parsers
import os
import tempfile
import glob

class TestVerilogComposer(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.dir_of_verilog_netlists = os.path.join(sdn.base_dir, "support_files", "verilog_netlists")
        cls.verilog_files = sorted(glob.glob(os.path.join(cls.dir_of_verilog_netlists, "*.v.zip")), key = os.path.getsize)

    @unittest.skip("Test takes a long time right now.")
    def test_large_verilog_compose(self):
        i = 0
        errors = 0
        for ii, filename in enumerate(self.verilog_files):
            with self.subTest(i=ii):
                if os.path.getsize(filename) <= 1024 * 10:
                    continue
                if filename.endswith(".zip"):
                    with tempfile.TemporaryDirectory() as tempdirectory:
                        # try:
                            print("*********************"+filename+"*********************")
                            # vp = sdn.parsers.verilog.parser.VerilogParser.from_filename(os.path.join(directory, filename))
                            # netlist = vp.parse()
                            netlist = parsers.parse(filename)
                            composers.compose(netlist, os.path.join(tempdirectory, os.path.basename(filename) +  "-spydrnet.v"))
                            #comp.run(netlist,"temp2/"+filename[:len(filename)-6] + "-spydrnet.v")
                            # comp.run(netlist,os.path.join(tempdirectory, filename[:len(filename)-6] + "-spydrnet.v"))
                            i+=1
                            print("pass")
                        # except Exception as identifier:
                        #     print("FAIL")
                        #     print(identifier)
                        #     errors += 1
                else:
                    continue

        print("processed",i,"errors", errors)
        
        assert errors == 0, "there were errors while parsing and composing files. Please see the output."

    def test_small_verilog_compose(self):
        i = 0
        errors = 0
        for ii, filename in enumerate(self.verilog_files):
            with self.subTest(i=ii):
                if os.path.getsize(filename) > 1024 * 10:
                    continue
                if filename.endswith(".zip"):
                    with tempfile.TemporaryDirectory() as tempdirectory:
                        # try:
                            print("*********************"+filename+"*********************")
                            # vp = sdn.parsers.verilog.parser.VerilogParser.from_filename(os.path.join(directory, filename))
                            # netlist = vp.parse()
                            netlist = parsers.parse(filename)
                            composers.compose(netlist, os.path.join(tempdirectory, os.path.basename(filename) +  "-spydrnet.v"))
                            #comp.run(netlist,"temp2/"+filename[:len(filename)-6] + "-spydrnet.v")
                            # comp.run(netlist,os.path.join(tempdirectory, filename[:len(filename)-6] + "-spydrnet.v"))
                            i+=1
                            print("pass")
                        # except Exception as identifier:
                        #     print("FAIL")
                        #     print(identifier)
                        #     errors += 1
                else:
                    continue

        print("processed",i,"errors", errors)
        
        assert errors == 0, "there were errors while parsing and composing files. Please see the output."

    def test_definition_list_option(self):
        for filename in glob.glob(os.path.join(
                self.dir_of_verilog_netlists, "*4bitadder.v.zip")):
            with tempfile.TemporaryDirectory() as tempdirectory:
                netlist = parsers.parse(filename)
                out_file = os.path.join(
                    tempdirectory, os.path.basename(filename) + "-spydrnet.v")
                composers.compose(netlist, out_file, definition_list=['adder'])

                with open(out_file, "r") as fp:
                    lines = fp.readlines()
                    print(len(lines))
                    m = list(filter(lambda x: x.startswith('module'), lines))
            self.assertGreater(len(m), 0, "Adder module not written")
            self.assertLess(len(m), 2, "Failed to write only definition_list")
            return
        raise AssertionError("Adder design not found " +
                             "definition_list options not tested,")

    def test_write_blackbox_option(self):
        for filename in glob.glob(os.path.join(
                self.dir_of_verilog_netlists, "*4bitadder.v.zip")):
            with tempfile.TemporaryDirectory() as tempdirectory:
                netlist = parsers.parse(filename)
                out_file = os.path.join(
                    tempdirectory, os.path.basename(filename) + "-spydrnet.v")
                composers.compose(netlist, out_file, write_blackbox=False)

                with open(out_file, "r") as fp:
                    lines = fp.readlines()
                    print(len(lines))
                    m = list(filter(lambda x: x.startswith('module'), lines))
            self.assertGreater(len(m), 0, "Adder module not written")
            self.assertLess(len(m), 2, "Failed to write only definition_list" +
                            "%s" % m)
            return
        raise AssertionError("definition_list options not test," +
                             "Adder design not found")



from spydrnet import Element, FirstClassElement, InnerPin, OuterPin, Wire, Netlist, Library, Definition, Instance,\
    Port, Cable
from spydrnet.global_state.global_service import lookup
from spydrnet.util.hierarchical_reference import HRef
from spydrnet.util.patterns import _is_pattern_absolute, _value_matches_pattern
from spydrnet.util.selection import Selection


def get_instances(obj, *args, **kwargs):
    """
    get_instances(obj, ...)

    Get instances *within* an object.

    Parameters
    ----------
    obj : object, Iterable - required
        The object or objects associated with this query. Queries return a collection objects associated with the
        provided object or objects that match the query criteria. For example, `sdn.get_instances(definition, ...)` would
        return all of the instances *within* the provided definition that match the additional criteria.
    patterns : str, Iterable - optional, positional or named, default: wildcard
        The search patterns. Patterns can be a single string or an Iterable collection of strings. Patterns can be
        absolute or they can contain wildcards or regular expressions. If `patterns` is not provided, then it defaults
        to a wildcard. Patterns are queried against the object property value stored under a specified key. Fast lookups
        are only attempted on absolute patterns that are not regular expressions and contain no wildcards.
    key : str, optional, default: ".NAME"
        This is the key that controls which value is being searched.
    is_case : bool - optional, named, default: True
        Specify if patterns should be treated as case sensitive. Only applies to patterns. Does not alter fast lookup
        behavior (if namespace policy uses case insensitive indexing, this parameter will not prevent a fast lookup
        from returning a matching object even if the case is not an exact match).
    is_re: bool - optional, named, default: False
        Specify if patterns are regular expressions. If `False`, a pattern can still contain `*` and `?` wildcards. A
        `*` matches zero or more characters. A `?` matches upto a single character.
    selection : Selection.{INSIDE, OUTSIDE}, default: INSIDE
        This parameter determines the instances that are returned based on the definition or instance that is being
        searched. This parameter only applies to objects that are definitions. If the selection is "INSIDE" (default),
        then the function will return all of the instances that are inside the definition (i.e., the definition's
        children) that match the remainder of the search criteria. If the selection is "OUTSIDE", then the function will
        return all of the instances of the provided definition that match the remainder of the search criteria.
    recursive : bool - optional, default: False
        Specify if search should be recursive or not meaning that sub hierarchical instances within an instance are
        included or not.
    filter : function
        This is a single input function that can be used to filter out unwanted virtual instances. If not specifed, all
        matching virtual instances are returned. Otherwise, virtual instances that cause the filter function to evaluate
        to true are the only items returned.
    
    Returns
    -------
    cables : generator
        The cables associated with a particular object or collection of objects.
    
    """
    # Check argument list
    if len(args) == 1 and 'patterns' in kwargs:
        raise TypeError("get_cables() got multiple values for argument 'patterns'")
    if len(args) > 1 or any(x not in {'patterns', 'key', 'filter', 'is_case', 'is_re', 'selection', 'recursive'} for x
                            in kwargs):
        raise TypeError("Unknown usage. Please see help for more information.")

    # Default values
    selection = kwargs.get('selection', Selection.INSIDE)
    if isinstance(selection, str):
        if selection in Selection.__members__:
            selection = Selection[selection]
    if selection not in {Selection.INSIDE, Selection.OUTSIDE}:
        raise TypeError("selection must be '{}'".format("', '".join([Selection.INSIDE.name, Selection.OUTSIDE.name])))

    filter_func = kwargs.get('filter', lambda x: True)
    is_case = kwargs.get('is_case', True)
    is_re = kwargs.get('is_re', False)
    patterns = args[0] if len(args) == 1 else kwargs.get('patterns', ".*" if is_re else "*")
    key = kwargs.get('key', ".NAME")
    recursive = kwargs.get('recursive', False)

    if isinstance(obj, (FirstClassElement, InnerPin, OuterPin, Wire)) is False:
        try:
            object_collection = list(iter(obj))
        except TypeError:
            object_collection = [obj]
    else:
        object_collection = [obj]
    if all(isinstance(x, (Element, HRef)) for x in object_collection) is False:
        raise TypeError("get_instances() only supports netlists, libraries, and definitions, or a collection of these "
                        "as the object searched")

    if isinstance(patterns, str):
        patterns = (patterns,)

    return _get_instances(object_collection, patterns, key, is_case, is_re, selection, recursive, filter_func)


def _get_instances(object_collection, patterns, key, is_case, is_re, selection, recursive, filter_func):
    for result in filter(filter_func, _get_instances_raw(object_collection, patterns, key, is_case, is_re, selection,
                                                         recursive)):
        yield result


def _get_instances_raw(object_collection, patterns, key, is_case, is_re, selection, recursive):
    found = set()
    other_instances = list()
    while object_collection:
        obj = object_collection.pop()
        if isinstance(obj, Definition):
            if selection == Selection.INSIDE:
                for pattern in patterns:
                    pattern_is_absolute = _is_pattern_absolute(pattern, is_case, is_re)
                    if pattern_is_absolute:
                        result = lookup(obj, Instance, key, pattern)
                        if result is not None and result not in found:
                            found.add(result)
                            yield result
                    else:
                        for instance in obj.children:
                            if key in instance:
                                value = instance[key] if key in instance else ''
                                if _value_matches_pattern(value, pattern, is_case, is_re) and instance not in found:
                                    found.add(instance)
                                    yield instance
                if recursive:
                    for child in obj.children:
                        reference = child.reference
                        if reference and reference.is_leaf() is False:
                            object_collection.append(reference)
            else:
                other_instances += obj.references
                if recursive:
                    for instance in obj.references:
                        parent = instance.parent
                        if parent:
                            object_collection.append(parent)
                        else:
                            other_instances.append(instance)
        elif isinstance(obj, Netlist):
            for library in obj.libraries:
                object_collection += library.definitions
        elif isinstance(obj, Library):
            object_collection += obj.definitions
        elif isinstance(obj, Instance):
            if selection is Selection.INSIDE:
                reference = obj.reference
                if reference:
                    other_instances += reference.children
                    if recursive:
                        object_collection += reference.children
            else:
                parent = obj.parent
                if parent:
                    other_instances += parent.references
                    if recursive:
                        object_collection += parent.references
        elif isinstance(obj, (Port, Cable)):
            definition = obj.definition
            if definition:
                other_instances += definition.references
        elif isinstance(obj, InnerPin):
            port = obj.port
            if port:
                object_collection.append(port)
        elif isinstance(obj, OuterPin):
            instance = obj.instance
            if instance:
                other_instances.append(instance)
        elif isinstance(obj, Wire):
            cable = obj.cable
            if cable:
                object_collection.append(cable)
        elif isinstance(obj, HRef):
            if obj.is_valid:
                item = obj.item
                if isinstance(item, Instance):
                    other_instances.append(item)
                else:
                    object_collection.append(item)

    if other_instances:
        namemap = dict()
        for other_instance in other_instances:
            if other_instance in found:
                continue
            found.add(other_instance)
            name = other_instance[key] if key in other_instance else ''
            if name not in namemap:
                namemap[name] = list()
            namemap[name].append(other_instance)
        for pattern in patterns:
            pattern_is_absolute = _is_pattern_absolute(pattern, is_case, is_re)
            if pattern_is_absolute:
                if pattern in namemap:
                    result = namemap[pattern]
                    for instance in result:
                        yield instance
            else:
                discard = set()
                for instance in found:
                    value = instance[key] if key in instance else ''
                    if _value_matches_pattern(value, pattern, is_case, is_re):
                        discard.add(instance)
                        yield instance
                found -= discard



# from spydrnet.ir.first_class_element import FirstClassElement
# from spydrnet.ir.instance import Instance
# from spydrnet.ir.definition import Definition
# from spydrnet.ir.netlist import Netlist
import unittest


from spydrnet.ir.first_class_element import FirstClassElement
from spydrnet.ir.instance import Instance
from spydrnet.ir.definition import Definition
from spydrnet.ir.netlist import Netlist
from spydrnet.ir.cable import Cable
from spydrnet.ir.wire import Wire
from spydrnet.util.hierarchical_reference import HRef
from spydrnet.ir.library import Library
from spydrnet.ir.port import Port

class TestShortcuts(unittest.TestCase):
    def test_hRef_shortcut(self):
        item = Instance("myCable")
        def2 = Definition("Hello")
        item.reference = def2
        hr = HRef(item)
        self.assertEqual(hr.item.name,item.name,'Href item shorcut error') 


    def test_netlist_top_instance_definition_shortcut(self):
        top_definition = Definition('this is my name')
        netlist = Netlist()
        netlist.top_instance = top_definition
        self.assertEqual(netlist.top_instance.reference.name,top_definition.name,'Netlist\'s top instance\'s shorcut error') 

    def test_if_leaf_shortcut(self):
        instance = Instance()
        definition = Definition()
        instance.reference = definition
        self.assertEqual(definition.is_leaf(), instance.is_leaf(), 'is_leaf shortcut error')
        definition.create_cable()
        instance.reference = definition
        self.assertEqual(definition.is_leaf(), instance.is_leaf(), 'is_leaf shortcut error')

    def test_wire_index_shortcut(self):
        cable = Cable()
        cable.create_wires(3)
        wire = Wire()
        cable.add_wire(wire,1)
        self.assertEqual(wire.cable.wires.index(wire), wire.index(), 'wire index shorcut error')
        self.assertEqual(wire.index(), 1, 'wire index shorcut error')
        wire2 = Wire()
        cable.add_wire(wire2,3)
        self.assertEqual(wire2.cable.wires.index(wire2), wire2.index(), 'wire index shorcut error')
        self.assertEqual(wire2.index(), 3, 'wire index shorcut error')

    def test_name_init_shortcut(self):
        d = Definition('d_name')
        self.assertEqual(d.name, 'd_name', 'Definition name init shorcut error')
        l = Library('l_name')
        self.assertEqual(l.name, 'l_name', 'Library name init shorcut error')
        n = Netlist('n_name')
        self.assertEqual(n.name, 'n_name', 'Netlist name init shorcut error')
        i = Instance('i_name')
        self.assertEqual(i.name, 'i_name', 'Instance name init shorcut error')
        c = Cable('c_name')
        self.assertEqual(c.name, 'c_name', 'Cable name init shorcut error')
        p = Port('p_name')
        self.assertEqual(p.name, 'p_name', 'Port name init shorcut error')

    def test_cable_shortcut(self):
        c = Cable('c_name', None,  False, True, 2)
        self.assertEqual(c.name, 'c_name', 'Cable name init shorcut error')
        self.assertEqual(c.is_downto, False, 'Cable is_downto init shorcut error')
        self.assertEqual(c.is_scalar, True, 'Cable is_scalar init shorcut error')
        self.assertEqual(c.lower_index, 2, 'Cable lower_index init shorcut error')

        c2 = Cable(is_scalar =  False)
        self.assertEqual(c2.name, None, 'Cable name init shorcut error')
        self.assertEqual(c2.is_downto, True, 'Cable is_downto init shorcut error')
        self.assertEqual(c2.is_scalar, False, 'Cable is_scalar init shorcut error')
        self.assertEqual(c2.lower_index, 0, 'Cable lower_index init shorcut error')


    def test_port_shortcut(self):
        p = Port('p_name', None, False, True, 4)

        self.assertEqual(p.name, 'p_name', 'Port name init shorcut error')
        self.assertEqual(p.is_downto, False, 'Port is_downto init shorcut error')
        self.assertEqual(p.is_scalar, True, 'Port is_scalar init shorcut error')
        self.assertEqual(p.lower_index, 4, 'Port lower_index init shorcut error')

        p2 = Port(is_downto =  False)
        self.assertEqual(p2.name, None, 'Port name init shorcut error')
        self.assertEqual(p2.is_downto, False, 'Port is_downto init shorcut error')
        self.assertEqual(p2.is_scalar, True, 'Port is_scalar init shorcut error')
        self.assertEqual(p2.lower_index, 0, 'Port lower_index init shorcut error')

    def test_properties_init_shortcut(self):
        
        d = Definition('d_name',{'key1':1, 'key2':'value2' })
        self.assertEqual(d['key1'], 1, 'Definition properties init shorcut error')

        l = Library('l_name',{'key1':'value1', 'key2':'value2' })
        self.assertEqual(l['key1'], 'value1', 'Library properties init shorcut error')

        n = Netlist('n_name',{'key1':'value1', 'key2': 63 })
        self.assertEqual(n['key2'], 63, 'Netlist properties init shorcut error')

        i = Instance('i_name',{'key1':'value1', 'key2':'value2' })
        self.assertEqual(i['key1'], 'value1', 'Instance properties init shorcut error')

        c = Cable('c_name',{'key1':'value1', 'key2':'value2' })
        self.assertEqual(c['key2'], 'value2', 'Cable properties init shorcut error')

        p = Port('p_name',{'key1':'value1', 'key2':'value2' })
        self.assertEqual(p['key1'], 'value1', 'Port properties init shorcut error')

    def test_library_child_instance_creation(self):
        l = Library('l_name',{'key1':'value1', 'key2':'value2' })
        self.assertEqual(l['key1'], 'value1', 'Library properties init shorcut error')

        l.create_definition("l_d_name", {'key1': 50, 'key2':'value2' })
        self.assertEqual('l_d_name', l.definitions[0].name, 'library\'s definition cretion shorcut error')
        self.assertEqual(l.definitions[0]['key1'], 50, 'library\'s definition cretion shorcut error')

        l2 = Library()
        l2.create_definition(properties = {'key1': 50, 'key2':'value2' })
        self.assertEqual(l2.definitions[0]['key1'], 50, 'library\'s definition cretion shorcut error')

    def test_definition_child_instance_creation(self):
        d = Definition('d_name',{'key1':1, 'key2':'value2' })
        self.assertEqual(d['key1'], 1, 'Definition properties init shorcut error')

        d.create_port('p_name')
        self.assertEqual(d.ports[0].name, 'p_name', 'Port name init shorcut error')
        d.create_port(properties = {'key1':'value1', 'key2':'value2' })
        self.assertEqual(d.ports[1]['key1'], 'value1', 'Port properties init shorcut error')

        d.create_port(properties = {'key1':'value1', 'key2':'value2' })
        self.assertEqual(d.ports[1]['key1'], 'value1', 'Port properties init shorcut error')

        d.create_child('d_c_name', {'key1':1, 'key2':'value2' })
        self.assertEqual(d.children[0]['key1'], 1, 'Definition properties init shorcut error')
        self.assertEqual(d['key1'], 1, 'Definition properties init shorcut error')

        d.create_cable('c_name', None,  False, True, 2)
        self.assertEqual(d.cables[0].name, 'c_name', 'Cable name init shorcut error')
        self.assertEqual(d.cables[0].is_downto, False, 'Cable is_downto init shorcut error')
        self.assertEqual(d.cables[0].is_scalar, True, 'Cable is_scalar init shorcut error')
        self.assertEqual(d.cables[0].lower_index, 2, 'Cable lower_index init shorcut error')

        d.create_cable(is_scalar =  False)
        self.assertEqual(d.cables[1].name, None, 'Cable name init shorcut error')
        self.assertEqual(d.cables[1].is_downto, True, 'Cable is_downto init shorcut error')
        self.assertEqual(d.cables[1].is_scalar, False, 'Cable is_scalar init shorcut error')
        self.assertEqual(d.cables[1].lower_index, 0, 'Cable lower_index init shorcut error')

    def test_netlist_child_instance_creation(self):
        n = Netlist('n_name')
        self.assertEqual(n.name, 'n_name', 'Netlist name init shorcut error')
        n.create_library('l_name',{'key1':'value1', 'key2':'value2' })
        self.assertEqual(n.libraries[0]['key1'], 'value1', 'Library properties init shorcut error')











import fnmatch
import re


def _is_pattern_absolute(pattern, is_case, is_re):
    if is_case is False or is_re is True:
        return False
    else:
        for char in pattern:
            if char in {"*", "?"}:
                return False
    return True


def _value_matches_pattern(value, pattern, is_case, is_re):
    if value is None:
        value = ""
    if is_re:
        try:
            if re.fullmatch(pattern, value, flags=0 if is_case else re.IGNORECASE):
                return True
        except re.error:
            return False
    elif is_case:
        return fnmatch.fnmatchcase(value, pattern)
    else:
        return fnmatch.fnmatch(value, pattern)


import unittest

from spydrnet.parsers.verilog.tokenizer import *
import spydrnet as sdn
from io import StringIO


class TestVerilogTokenizer(unittest.TestCase):


    def test_against_4bit_adder(self):
        directory = os.path.join(sdn.base_dir, "support_files", "verilog_netlists", "4bitadder.v.zip")
        tokenizer = VerilogTokenizer.from_filename(directory)
        while(tokenizer.has_next()):
            #print(tokenizer.next())
            tokenizer.next()

    def test_spaces(self):
        tokenizer = VerilogTokenizer.from_string("wire temp = 1'b1; if something == some2")
        while(tokenizer.has_next()):
            #print(tokenizer.next())
            tokenizer.next()
        tokenizer = VerilogTokenizer.from_string("wire temp=1'b1 ;if something==some2")
        while(tokenizer.has_next()):
            #print(tokenizer.next())
            tokenizer.next()

    def test_from_stream(self):
        stream = StringIO("module empty();endmodule")
        tokenizer = VerilogTokenizer.from_stream(stream)
        expected_tokens = ["module", "empty", "(", ")", ";", "endmodule"]
        tokens = []
        while(tokenizer.has_next()):
            tokens.append(tokenizer.next())
        
        for i in range(len(expected_tokens)):
            assert expected_tokens[i] == tokens[i]


import unittest
import spydrnet as sdn


class TestGetPorts(unittest.TestCase):
    def test_parameter_checking(self):
        definition = sdn.Definition()
        port = definition.create_port()
        port.name = "MY_PORT"
        self.assertRaises(TypeError, sdn.get_ports, definition, "MY_PORT", patterns="MY_PORT")
        self.assertRaises(TypeError, sdn.get_ports, definition, "MY_PORT", unsupported_keyword=None)
        self.assertRaises(TypeError, sdn.get_ports, None, "MY_PORT")
        self.assertRaises(TypeError, sdn.get_ports, [None, definition], "MY_PORT")

    def test_collection(self):
        definition = sdn.Definition()
        port = definition.create_port()
        port.name = "MY_PORT"
        instance = sdn.Instance()
        instance.name = "MY_INST"
        ports = list(sdn.get_ports([definition, instance]))
        self.assertEqual(len(ports), 1)

    def test_get_ports_on_instance(self):
        definition = sdn.Definition()
        port = definition.create_port()
        port.name = "MY_PORT"
        instance = sdn.Instance()
        instance.reference = definition
        port1 = next(instance.get_ports("MY_PORT"))
        self.assertEqual(port, port1)

    def test_get_ports_in_library(self):
        library = sdn.Library()
        definition = library.create_definition()
        port = definition.create_port()
        port.name = "MY_PORT"
        instance = sdn.Instance()
        instance.reference = definition
        port1 = next(library.get_ports("MY_PORT"))
        self.assertEqual(port, port1)

    def test_get_ports_in_netlist(self):
        netlist = sdn.Netlist()
        library = netlist.create_library()
        definition = library.create_definition()
        port = definition.create_port()
        port.name = "MY_PORT"
        instance = sdn.Instance()
        instance.reference = definition
        port1 = next(netlist.get_ports("MY_PORT"))
        self.assertEqual(port, port1)

    def test_get_port_reflection(self):
        port = sdn.Port()
        port.name = "MY_PORT"
        search = next(sdn.get_ports(port, "MY_PORT"))
        self.assertEqual(port, search)

    def test_get_port_inner_pin(self):
        port = sdn.Port()
        pin = port.create_pin()
        search = next(sdn.get_ports(pin))
        self.assertEqual(port, search)
        port.remove_pin(pin)
        search = next(sdn.get_ports(pin), None)
        self.assertIsNone(search)

    def test_get_port_instance_and_outer_pin(self):
        definition = sdn.Definition()
        port = definition.create_port()
        pin = port.create_pin()
        instance = sdn.Instance()
        instance.reference = definition

        search = next(sdn.get_ports(instance))
        self.assertIs(port, search)

        outer_pin = instance.pins[pin]
        search = next(sdn.get_ports(outer_pin))
        self.assertIs(port, search)

    def test_get_ports_href_cable(self):
        netlist = sdn.Netlist()
        library = netlist.create_library()
        definition = library.create_definition()
        port = definition.create_port()
        pin = port.create_pin()
        cable = definition.create_cable()
        wire = cable.create_wire()
        wire.connect_pin(pin)
        instance = sdn.Instance()
        instance.reference = definition
        netlist.top_instance = instance
        href = next(sdn.get_hcables(cable))

        search = next(sdn.get_ports(href))
        self.assertIs(port, search)

    def test_unique(self):
        netlist = sdn.Netlist()
        library = netlist.create_library()
        definition = library.create_definition()
        port = definition.create_port()
        pin = port.create_pin()
        cable = definition.create_cable()
        wire = cable.create_wire()
        wire.connect_pin(pin)
        instance = sdn.Instance()
        instance.reference = definition

        search = list(sdn.get_ports([netlist, cable]))
        self.assertIs(port, search[0])

        search = list(sdn.get_ports(cable))
        self.assertIs(port, search[0])



from spydrnet.ir.bundle import Bundle
from spydrnet.ir.wire import Wire
from spydrnet.ir.views.listview import ListView
from spydrnet.global_state import global_callback
from spydrnet.global_state.global_callback import _call_create_cable
from copy import deepcopy, copy, error


class Cable(Bundle):
    """Representation of several wires in a collection.

    Much like ports, cables extend the bundle class, giving them indexing ability. They represent several wires in a collection or bus that are generally related.
    This could be thought of much like vector types in VHDL ie std_logic_vector (7 downto 0)"""
    __slots__ = ['_wires']

    def __init__(self, name=None, properties=None, is_downto=None, is_scalar=None, lower_index=None):
        """Create a cable with no wires and default values for a bundle.

        parameters
        ----------

        name - (str) the name of this instance
        properties - (dict) the dictionary which holds the properties
        id_downto - (bool) set the downto status. Downto is False if the right index is higher than the left one, True otherwise
        is_scalar - (bool) set the scalar status. Return True if the item is a scalar False otherwise.
        lower_index - (int) get the value of the lower index of the array.

        """
        super().__init__()
        self._wires = list()
        _call_create_cable(self)
        if name != None:
            self.name = name

        if is_downto is not None:
            self.is_downto = is_downto

        if is_scalar is not None:
            self.is_scalar = is_scalar

        if lower_index is not None:
            self.lower_index = lower_index

        if properties != None:
            assert isinstance(
                properties, dict), "properties must be a dictionary"
            for key in properties:
                self[key] = properties[key]

    def _items(self):
        """Overrides the bundle _items function to return wires"""
        return self._wires

    @property
    def wires(self):
        """Gets a list of wires that are in this cable"""
        return ListView(self._wires)

    @wires.setter
    def wires(self, value):
        """Sets the wires to a reordered list of wires.

        This function is to be used for reordering of wires
        """
        value_list = list(value)
        value_set = set(value_list)
        # try:
        assert len(value_list) == len(value_set) and set(self._wires) == value_set, \
            "Set of values does not match, assigment can only be used for reordering, values must be unique"
        # except:
        #     import pdb; pdb.set_trace()
        self._wires = value_list

    def create_wires(self, wire_count):
        """Creates wire_count wires for this cable and adds them to it.

        parameters
        ----------

        wire_count - (int) the number of wires to be added to the cable.
        """
        for _ in range(wire_count):
            self.create_wire()
        return self.wires[-wire_count:]

    def create_wire(self):
        """Creates a wire and adds it to the cable. Returns the wire that was created"""
        wire = Wire()
        self.add_wire(wire)
        return wire

    def add_wire(self, wire, position=None):
        """Adds a wire to the cable at the given position. This wire must not belong to a cable already

        parameters
        ----------

        wire - (Wire) the wire to be added to the cable. This wire must not belong to any other cable.

        position - (int, default None) the index in the wires list at which to add the wire.
        """
        assert wire.cable is not self, "Wire already belongs to this cable"
        assert wire.cable is None, "Wire already belongs to a different cable"
        global_callback._call_cable_add_wire(self, wire)
        if position is not None:
            self._wires.insert(position, wire)
        else:
            self._wires.append(wire)
        wire._cable = self

    def remove_wire(self, wire):
        """removes the given wire from the cable and return it. The wire must belong to this cable

        parameters
        ----------

        wire - (Wire) the wire to be removed from the cable.
        """
        assert wire.cable == self, "Wire does not belong to this cable"
        self._remove_wire(wire)
        self._wires.remove(wire)

    def remove_wires_from(self, wires):
        """Remove all wires given from the cable. Each must be a member of this cable.

        parameters
        ----------

        wires - (List of Wire objects) wires to be removed from the cable.
        """
        if isinstance(wires, set):
            excluded_wires = wires
        else:
            excluded_wires = set(wires)
        assert all(
            x.cable == self for x in excluded_wires), "Some wires do not belong to this cable"
        for wire in excluded_wires:
            self._remove_wire(wire)
        self._wires = list(x for x in self._wires if x not in excluded_wires)

    def _remove_wire(self, wire):
        """Internal wire removal call. dissociates the wire from the cable"""
        global_callback._call_cable_remove_wire(self, wire)
        wire._cable = None

    def _clone_rip_and_replace(self, memo):
        """Remove from its current environment and place it into the new cloned environment with references held in the memo dictionary"""
        for w in self._wires:
            w._clone_rip_and_replace(memo)

    def _clone_rip(self):
        """Remove from its current environmnet. This will remove all pin pointers and create a floating stand alone instance."""
        for w in self._wires:
            w._clone_rip()
            w._cable = self

    def _clone(self, memo):
        """Not api safe clone function
        clone leaving all references in tact.
        the element can then either be ripped or ripped and replaced"""
        assert self not in memo, "the object should not have been copied twice in this pass"
        c = Cable()
        memo[self] = c
        new_wires = list()
        for w in self._wires:
            new_wires.append(w._clone(memo))
        c._wires = new_wires
        for w in c._wires:
            w._cable = c
        c._definition = None
        c._is_downto = deepcopy(self._is_downto)
        c._is_scalar = deepcopy(self._is_scalar)
        c._lower_index = deepcopy(self._lower_index)
        c._data = deepcopy(self._data)
        return c

    def clone(self):
        """
        Clone the Cable and all of its wires in an api safe way
        the following will be true of the returned cable

         * The cable will be orphaned from any definitions
         * the wires in the cable will not be connected to any pins
         * is_downto, is_scalar, lower_index will be maintained
         * the wires in the cable will all have the cable set as the parent
        """
        c = self._clone(dict())
        c._clone_rip()
        return c

    def __str__(self):
        """Re-define the print function so it is easier to read"""
        rep = str(type(self))
        rep = rep[:-1] + '; '
        if self.is_downto is not None and self.is_downto is True:
            rep += 'is_downto: True; '
        else:
            rep += 'is_downto: False; '
        if self.is_scalar is True:
            rep += 'is_scalar: True; '
        else:
            rep += 'is_scalar: False; '
        rep += 'lower index: ' + str(self.lower_index)
        rep += '>'
        return rep




import unittest
from spydrnet import Netlist, Instance
from spydrnet.uniquify import uniquify

class TestUniquify(unittest.TestCase):
    def create_netlist(self):
        nl = Netlist()
        lib = nl.create_library()
        d1 = lib.create_definition()
        d2 = lib.create_definition()
        d3 = lib.create_definition(name='instance_name')
        i11 = d1.create_child()
        i12 = d1.create_child()
        i13 = d1.create_child()
        i24 = d2.create_child()
        i25 = d2.create_child()
        i26 = d2.create_child()
        i27 = d2.create_child()
        d4 = lib.create_definition()
        i38 = d3.create_child()
        i39 = d3.create_child()
        i11.reference = d3
        i12.reference = d3
        i13.reference = d3
        i24.reference = d1
        i25.reference = d1
        i26.reference = d3
        i27.reference = d3
        i38.reference = d4
        i39.reference = d4
        nl.top_instance = Instance()
        nl.top_instance.reference = d2
        return nl

    def is_unique(self, netlist):
        #get all the definitions that are instanced below the top instance
        top_inst = netlist.top_instance
        top_def = top_inst.reference
        inst = []
        inst.append(top_inst)
        index = 1
        for c in top_def.children:
            inst.append(c)
        while index != len(inst):
            for c in inst[index].reference.children:
                inst.append(c)
            index += 1

        #now we have a list of all the instances in the netlist make sure all references are unique to the netlist
        refs = set()
        for i in inst:
            if i.reference.is_leaf():
                continue
            if i.reference in refs:
                return False #the instance references a previously referenced definition, instances are not unique
            refs.add(i.reference)
        return True

    def test_already_unique(self):
        nl = Netlist()
        lib = nl.create_library()
        d1 = lib.create_definition()
        d2 = lib.create_definition()
        d3 = lib.create_definition()
        i1 = d1.create_child()
        i2 = d1.create_child()
        i1.reference = d2
        i2.reference = d3
        top = Instance()
        top.reference = d1
        nl.top_instance = top
        assert self.is_unique(nl), "netlist should be unique upon creation in this test"
        uniquify(nl)
        assert self.is_unique(nl), "netlist should remain unique in this test. somehow uniquify made the netlist un-unique"

    # def simple_recursive_netlist_visualizer(self, netlist):
    #     #TODO put this code somewhere where people can use it to visualize simple netlists
    #     top_instance = netlist.top_instance
    #     #should look something like this:
    #     #top
    #     #   child1
    #     #       child1.child
    #     #   child2
    #     #       child2.child
    #     def recurse(instance, depth):
    #         s = depth * "\t"
    #         print(s, instance.name, "(", instance.reference.name, ")")
    #         for c in instance.reference.children:  
    #             recurse(c, depth + 1)
        
    #     recurse(top_instance, 0)


    def test_uniquify_simple_with_names(self):
        '''simple test with 2 definitions.'''
        nl = Netlist()
        lib = nl.create_library()
        d1 = lib.create_definition()
        d1.name = "definition1"
        d2 = lib.create_definition()
        d2.name = "definition2"
        d3 = lib.create_definition()
        d3.name = "leaf definition"
        i1 = d1.create_child()
        i1.name = "instance1"
        i2 = d1.create_child()
        i2.name = "instance2"
        i1.reference = d2
        i2.reference = d2
        i3 = d2.create_child()
        i3.name = "leaf cell"
        i3.reference = d3
        top = Instance()
        top.name = "top instance"
        top.reference = d1
        nl.top_instance = top
        assert not self.is_unique(nl), "initial netlist should not be unique in this test"
        uniquify(nl)
        assert self.is_unique(nl), "netlist should have been uniquified."

    def test_uniquify(self):
        nl = self.create_netlist()
        assert not self.is_unique(nl), "our test netlist should not be unique"
        uniquify(nl)
        assert self.is_unique(nl), "our netlist should have been uniquified"


from spydrnet.ir.first_class_element import FirstClassElement
from spydrnet.ir.library import Library
from spydrnet.ir.instance import Instance
from spydrnet.ir.views.listview import ListView
from spydrnet.global_state import global_callback
from spydrnet.global_state.global_callback import _call_create_netlist
from copy import deepcopy, copy, error
from spydrnet.ir.definition import Definition


class Netlist(FirstClassElement):
    """Represents a netlist object.

    Contains a top level instance and libraries

    Examples
    --------
    After importing the spydrnet package, we can initialize the netlist from scratch

    >>> import spydrnet as sdn
    >>> netlist = sdn.Netlist()

    We can also initalize an top instance for the netlist. For more info: :ref:`Instance`

    >>> top_instance = sdn.Instance()
    >>> netlist.top_instance = top_instance

    This is the way to set up a top level definition for the netlist

    >>> top_definition = sdn.Definition()
    >>> netlist.top_instance = top_definition

    We can initialize a "primatives" and a "work" library this way:

    >>> primitives_library = netlist.create_library()
    >>> primitives_library['EDIF.identifier'] = 'hdi_primitives'

    >>> work_library = netlist.create_library()
    >>> work_library['EDIF.identifier'] = 'work'
    """
    __slots__ = ['_libraries', '_top_instance']

    def __init__(self, name=None, properties=None):
        """
        creates an empty object of type netlist

        parameters
        ----------

        name - (str) the name of this instance
        properties - (dict) the dictionary which holds the properties
        """
        super().__init__()
        self._libraries = list()
        self._top_instance = None
        _call_create_netlist(self)

        if name != None:
            self.name = name
        if properties != None:
            assert isinstance(
                properties, dict), "properties must be a dictionary"
            for key in properties:
                self[key] = properties[key]

    def compose(self, *args, **kwargs):
        """Compose a netlist into a file format.

        Compose(filename).
        Shortcut to :func:`~spydrnet.compose`.
        """
        from spydrnet.composers import compose
        compose(self, *args, **kwargs)

    @property
    def libraries(self):
        """Get a list of all libraries included in the netlist."""
        return ListView(self._libraries)

    @libraries.setter
    def libraries(self, value):
        """Set the libraries.

        This function can only be used to reorder the libraries. Use the remove_library and
        add_library functions to add and remove libraries.

        Parameters
        ----------
        value - the reordered list of libraries

        """
        value_list = list(value)
        value_set = set(value_list)
        assert len(value_list) == len(value_set) and set(self._libraries) == value_set, \
            "Set of values do not match, this assignment can only reorder values, values must be unique"
        self._libraries = value_list

    @property
    def top_instance(self):
        """Get the top instance in the netlist.

        Returns
        -------
        Instance
            The top level instance in the environment

        """
        return self._top_instance

    @top_instance.setter
    def top_instance(self, instance):
        """Sets the top instance of the design.

        The instance must not be null and should probably come from this netlist

        Parameters
        ----------
        instance - (Instance or Definition) the instance to set as the top instance. If a definition is passed into the funciton,
        creates a new instance with that definition and set it as the top instance.
        """
        assert instance is None or isinstance(instance, Instance) or isinstance(
            instance, Definition), "Must specify an instance"
        global_callback._call_netlist_top_instance(self, instance)
        # TODO: should We have a DRC that makes sure the instance is of a definition contained in netlist? I think no
        #  but I am open to hear other points of veiw.
        if self.top_instance:
            self.top_instance.is_top_instance = False
        if isinstance(instance, Definition):
            top = Instance()
            top.reference = instance
            top.is_top_instance = True
            self.top_instance = top
        else:
            self._top_instance = instance
            if instance:
                instance.is_top_instance = True

    def set_top_instance(self, instance, instance_name='instance'):
        """Sets the top instance of the design.

        The instance must not be null and should probably come from this netlist

        Parameters
        ----------
        instance - (Instance or Definition) the instance to set as the top instance. If a definition is passed into the funciton,
        creates a new instance with that definition and set it as the top instance.
        """
        assert instance is None or isinstance(instance, Instance) or isinstance(
            instance, Definition), "Must specify an instance"
        global_callback._call_netlist_top_instance(self, instance)
        # TODO: should We have a DRC that makes sure the instance is of a definition contained in netlist? I think no
        #  but I am open to hear other points of veiw.

        if isinstance(instance, Definition):
            top = Instance()
            top.reference = instance
            instance.name = instance_name
            self.top_instance = top
            self.top_instance.name = instance_name
        else:
            self._top_instance = instance

    def create_library(self, name=None, properties=None):
        """Create a library and add it to the netlist and return that library

        parameters
        ----------

        name - (str) the name of the library
        properties - (dict) the dictionary which holds the properties of the library
        """

        library = Library(name, properties)
        self.add_library(library)
        return library

    def add_library(self, library, position=None):
        """add an already existing library to the netlist. 

        This library should not belong to another netlist. Use
        remove_library from other netlists before adding

        Parameters
        ----------
        library - Library
            The library to be added to the netlist.
        position - int, (default None)
            When set, it is the index at which to add the library in the libraries list.

        """
        assert library not in self._libraries, "Library already included in netlist"
        assert library.netlist is None, "Library already belongs to a different netlist"
        global_callback._call_netlist_add_library(self, library)
        if position is not None:
            self._libraries.insert(position, library)
        else:
            self._libraries.append(library)
        library._netlist = self

    def remove_library(self, library):
        """Removes the given library if it is in the netlist

        Parameters
        ----------
        library - Library
            The library to be removed.

        """
        assert library.netlist == self, "Library is not included in netlist"
        self._remove_library(library)
        self._libraries.remove(library)

    def remove_libraries_from(self, libraries):
        """Removes all the given libraries from the netlist.

        All libraries must be in the netlist.

        Parameters
        ----------
        libraries - Set
            Libraries to be removed.

        """
        if isinstance(libraries, set):
            excluded_libraries = libraries
        else:
            excluded_libraries = set(libraries)
        assert all(x.netlist == self for x in excluded_libraries), "Some libraries to remove are not included in " \
                                                                   "netlist "
        included_libraries = list()
        for library in self._libraries:
            if library not in excluded_libraries:
                included_libraries.append(library)
            else:
                self._remove_library(library)
        self._libraries = included_libraries

    def _remove_library(self, library):
        """Internal function which will separate a particular libraries binding from the netlist."""
        global_callback._call_netlist_remove_library(self, library)
        library._netlist = None

    def _clone_rip(self, memo):
        """Need to remove any extraneous references to floating, no parent instances"""
        for lib in self._libraries:
            for defin in lib._definitions:
                new_ref = set()
                for ref in defin._references:
                    if ref in memo.values():
                        new_ref.add(ref)
                defin._references = new_ref

    def _clone(self, memo):
        """Clone leaving all references in tact.

        The element can then either be ripped or ripped and replaced.
        """
        assert self not in memo, "the object should not have been copied twice in this pass"
        c = Netlist()
        memo[self] = c
        c._data = deepcopy(self._data)

        new_libraries = list()
        for library in self._libraries:
            new_libraries.append(library._clone(memo))
        c._libraries = new_libraries

        if self._top_instance == None:
            c._top_instance = None
        else:
            if self._top_instance in memo:
                c._top_instance = memo[self._top_instance]
            else:
                new_top = self.top_instance._clone(memo)
                new_top._clone_rip_and_replace_in_definition(memo)
                new_top._clone_rip_and_replace_in_library(memo)
                c._top_instance = new_top

        for library in c._libraries:
            library._netlist = c
            library._clone_rip_and_replace(memo)
        return c

    def clone(self):
        """API safe clone on a netlist.

        This clone function should act just the way you would expect
        All references are internal to the netlist that has been cloned.
        """
        memo = dict()
        c = self._clone(memo)
        c._clone_rip(memo)
        return c

    def __str__(self):
        """Re-define the print function so it is easier to read"""
        rep = super().__str__()
        rep = rep[:-1] + '; '
        if self.top_instance is None:
            rep += 'top_instance undefined'
        elif self.top_instance.name is None:
            rep += 'top_instance.name undefined'
        else:
            rep += 'top_instance.name \'' + self.top_instance.name + '\''
        rep += '>'
        return rep

    def is_unique(self):
        """
        checks whether all of the instances in the netlist are unique or not
        """
        for instance in self.get_instances():
            if len(instance.reference.references) == 1 or instance.reference.is_leaf():
                continue
            else:
                return False
        return True



from spydrnet import FirstClassElement, InnerPin, OuterPin, Wire, Netlist, Library, Definition, Instance, Port, Cable
from spydrnet.util.hierarchical_reference import HRef
from spydrnet.util.patterns import _is_pattern_absolute, _value_matches_pattern


def get_hports(obj, *args, **kwargs):
    """
    get_hports(obj, ...)

    Get hierarchical references to ports *within* an object.

    Parameters
    ----------
    obj : object, Iterable - required
        The object or objects associated with this query. Queries return a collection of objects associated with the
        provided object or objects that match the query criteria. For example, `sdn.get_instances(netlist, ...)` would
        return all of the instances *within* the provided definition that match the additional criteria.
    patterns : str, Iterable - optional, positional or named, default: wildcard
        The search patterns. Patterns can be a single string or an Iterable collection of strings. Patterns can be
        absolute or they can contain wildcards or regular expressions. If `patterns` is not provided, then it defaults
        to a wildcard.
    recursive : bool - optional, default: False
        Specify if search should be recursive or not meaning that sub hierarchical pins within an instance are
        included or not.
    is_case : bool - optional, named, default: True
        Specify if patterns should be treated as case sensitive. Only applies to patterns. Does not alter fast lookup
        behavior (if namespace policy uses case insensitive indexing, this parameter will not prevent a fast lookup
        from returning a matching object even if the case is not an exact match).
    is_re: bool - optional, named, default: False
        Specify if patterns are regular expressions. If `False`, a pattern can still contain `*` and `?` wildcards. A
        `*` matches zero or more characters. A `?` matches upto a single character.
    filter : function
        This is a single input function that can be used to filter out unwanted virtual instances. If not specifed, all
        matching virtual instances are returned. Otherwise, virtual instances that cause the filter function to evaluate
        to true are the only items returned.
    
    Returns
    -------
    href_ports : generator
        The hierarchical references to ports associated with a particular object or collection of objects.
    
    """
    # Check argument list
    if len(args) == 1 and 'patterns' in kwargs:
        raise TypeError("get_hports() got multiple values for argument 'patterns'")
    if len(args) > 1 or any(x not in {'patterns', 'recursive', 'filter', 'is_case', 'is_re'} for x in
                            kwargs):
        raise TypeError("Unknown usage. Please see help for more information.")

    # Default values
    filter_func = kwargs.get('filter', lambda x: True)
    recursive = kwargs.get('recursive', False)
    is_case = kwargs.get('is_case', True)
    is_re = kwargs.get('is_re', False)
    patterns = args[0] if len(args) == 1 else kwargs.get('patterns', ".*" if is_re else "*")

    if isinstance(obj, (FirstClassElement, InnerPin, OuterPin, Wire)) is False:
        try:
            object_collection = list(iter(obj))
        except TypeError:
            object_collection = [obj]
    else:
        object_collection = [obj]
    if all(isinstance(x, (HRef, FirstClassElement, InnerPin, OuterPin, Wire)) for x in object_collection) is False:
        raise TypeError("get_hports() supports all netlist related objects and hierarchical references or a "
                        "collection of theses as the object searched, unsupported object provided")

    if isinstance(patterns, str):
        patterns = (patterns,)
    assert isinstance(patterns, (FirstClassElement, InnerPin, OuterPin, Wire)) is False

    return _get_hports(object_collection, patterns, recursive, is_case, is_re, filter_func)


def _get_hports(object_collection, patterns, recursive, is_case, is_re, filter_func):
    for result in filter(filter_func, _get_hports_raw(object_collection, patterns, recursive, is_case, is_re)):
        yield result


def _get_hports_raw(object_collection, patterns, recursive, is_case, is_re):
    in_namemap = set()
    in_yield = set()
    namemap = dict()
    bypass_namesearch = set()
    while object_collection:
        obj = object_collection.pop()
        if isinstance(obj, Netlist):
            top_instance = obj.top_instance
            if top_instance:
                href = HRef.from_parent_and_item(None, top_instance)
                object_collection.append(href)
        elif isinstance(obj, HRef):
            if obj.is_valid is False:
                continue
            item = obj.item
            if isinstance(item, Instance):
                if obj not in bypass_namesearch:
                    _update_hport_namemap(obj, recursive, in_namemap, namemap)
                else:
                    bypass_namesearch.discard(obj)
                    reference = item.reference
                    if reference:
                        # Get all cables inside a hierarchical instance
                        for port in reference.ports:
                            hport = HRef.from_parent_and_item(obj, port)
                            if hport not in in_yield:
                                in_yield.add(hport)
                                yield(hport)
                        # get internal cables recursively
                        if recursive:
                            for child in reference.children:
                                href_child = HRef.from_parent_and_item(obj, child)
                                bypass_namesearch.add(href_child)
                                object_collection.append(href_child)
            elif isinstance(item, Port):
                if obj not in in_yield:
                    in_yield.add(obj)
                    yield (obj)
            elif isinstance(item, Cable):
                for wire in item.wires:
                    href_wire = HRef.from_parent_and_item(obj, wire)
                    object_collection.append(href_wire)
            elif isinstance(item, Wire):
                    href_parent_cable = obj.parent
                    href_parent_instance = href_parent_cable.parent
                    for pin in item.pins:
                        if isinstance(pin, OuterPin):
                            instance = pin.instance
                            if instance:
                                href_inst = HRef.from_parent_and_item(href_parent_instance, pin.instance)
                                inner_pin = pin.inner_pin
                                if inner_pin:
                                    inner_port = inner_pin.port
                                    if inner_port:
                                        href_port = HRef.from_parent_and_item(href_inst, inner_port)
                                        if href_port not in in_yield:
                                            in_yield.add(href_port)
                                            yield href_port
                        else:
                            port = pin.port
                            if port:
                                href_port = HRef.from_parent_and_item(href_parent_instance, port)
                                if href_port not in in_yield:
                                    in_yield.add(href_port)
                                    yield href_port
            elif isinstance(item, InnerPin):
                hport = obj.parent
                if hport not in in_yield:
                    in_yield.add(hport)
                    yield hport
        elif isinstance(obj, Library):
            object_collection += obj.definitions
        elif isinstance(obj, Definition):
            hrefs = set(HRef.get_all_hrefs_of_instances(obj.references))
            bypass_namesearch |= hrefs
            object_collection += hrefs
        elif isinstance(obj, Instance):
            hrefs = set(HRef.get_all_hrefs_of_instances(obj))
            bypass_namesearch |= hrefs
            object_collection += hrefs
        elif isinstance(obj, (Port, Cable, InnerPin, OuterPin, Wire)):
            object_collection += HRef.get_all_hrefs_of_item(obj)

    for href in in_yield:
        in_namemap.discard(href)

    if in_namemap:  # namemap is to cable
        for pattern in patterns:
            pattern_is_absolute = _is_pattern_absolute(pattern, is_case, is_re)
            if pattern_is_absolute:
                if pattern in namemap:
                    result = namemap[pattern]
                    for href in result:
                        if href in in_namemap:
                            in_namemap.remove(href)
                            yield href
            else:
                for name in namemap:
                    if _value_matches_pattern(name, pattern, is_case, is_re):
                        result = namemap[name]
                        for href in result:
                            if href in in_namemap:
                                in_namemap.remove(href)
                                yield href


def _update_hport_namemap(href_instance, recursive, found, namemap):
    search_stack = [(href_instance, False)]
    name_stack = list()
    while search_stack:
        href_instance, visited = search_stack.pop()
        if visited:
            name_stack.pop()
        else:
            search_stack.append((href_instance, True))
            name_stack.append(href_instance.item.name if href_instance.item.name else '')
            item = href_instance.item
            reference = item.reference
            if reference:
                for port in reference.ports:
                    hport = HRef.from_parent_and_item(href_instance, port)
                    name_stack.append(port.name if port.name else '')
                    port_hname = '/'.join(name_stack[1:])
                    if hport not in found:
                        found.add(hport)
                        if port_hname not in namemap:
                            namemap[port_hname] = list()
                        namemap[port_hname].append(hport)
                    name_stack.pop()
                if recursive:
                    for child in reference.children:
                        if child.reference:
                            href_child = HRef.from_parent_and_item(href_instance, child)
                            search_stack.append((href_child, False))



from spydrnet.util.get_netlists import get_netlists
from spydrnet.util.get_libraries import get_libraries
from spydrnet.util.get_definitions import get_definitions
from spydrnet.util.get_ports import get_ports
from spydrnet.util.get_pins import get_pins
from spydrnet.util.get_cables import get_cables
from spydrnet.util.get_wires import get_wires
from spydrnet.util.get_instances import get_instances

from spydrnet.util.get_hinstances import get_hinstances
from spydrnet.util.get_hports import get_hports
from spydrnet.util.get_hpins import get_hpins
from spydrnet.util.get_hcables import get_hcables
from spydrnet.util.get_hwires import get_hwires

from spydrnet.util.hierarchical_reference import HRef



from spydrnet.ir import Element, InnerPin, OuterPin, Wire, Netlist, Library, Definition, Instance,\
    Port, Cable
from spydrnet.util.hierarchical_reference import HRef
from spydrnet.util.selection import Selection


def get_pins(obj, *args, **kwargs):
    """
    get_pins(obj, ...)

    Get pins *within* an object.

    Parameters
    ----------
    obj : object, Iterable - required
        The object or objects associated with this query. Queries return a collection objects associated with the
        provided object or objects that match the query criteria. For example, `sdn.get_ports(definition, ...)` would
        return all of the ports associated with the provided definition that match the additional criteria.
    selection : Selection.{INSIDE, OUTSIDE}, default: INSIDE
        Controls the type of pin returned. Setting this parameter to "OUTSIDE" will return the outer pins of the objects
        queried. For cables, this returns the corresponding pin based on this parameter.
    filter : function
        This is a single input function that can be used to filter out unwanted virtual instances. If not specifed, all
        matching virtual instances are returned. Otherwise, virtual instances that cause the filter function to evaluate
        to true are the only items returned.
    
    Returns
    -------
    pins : generator
       The pins associated with a particular object or collection of objects.
    
    """
    # Check argument list
    if len(args) > 0 or any(x not in {'filter', 'selection'} for x in kwargs):
        raise TypeError("Unknown usage. Please see help for more information.")

    # Default values
    selection = kwargs.get('selection', Selection.INSIDE)
    if isinstance(selection, str):
        if selection in Selection.__members__:
            selection = Selection[selection]
    if selection not in {Selection.INSIDE, Selection.OUTSIDE}:
        raise TypeError("selection must be '{}'".format("', '".join([Selection.INSIDE.name, Selection.OUTSIDE.name])))

    filter_func = kwargs.get('filter', lambda x: True)

    if isinstance(obj, (Element, HRef)) is False:
        try:
            object_collection = list(iter(obj))
        except TypeError:
            object_collection = [obj]
    else:
        object_collection = [obj]
    if all(isinstance(x, (Element, HRef)) for x in object_collection) is False:
        raise TypeError("get_ports() supports netlist elements and hierarchical references, or a collection of "
                        "these as the object searched")

    return _get_ports(object_collection, selection, filter_func)


def _get_ports(object_collection, selection, filter_func):
    for result in filter(filter_func, _get_ports_raw(object_collection, selection)):
        yield result


def _get_ports_raw(object_collection, selection):
    found = set()
    while object_collection:
        obj = object_collection.pop()
        if isinstance(obj, Definition):
            for port in obj.ports:
                object_collection += port.pins
        elif isinstance(obj, Netlist):
            for library in obj.libraries:
                for definition in library.definitions:
                    object_collection.append(definition)
        elif isinstance(obj, Library):
            object_collection += obj.definitions
        elif isinstance(obj, Instance):
            object_collection += obj.pins
        elif isinstance(obj, Port):
            object_collection += obj.pins
        elif isinstance(obj, InnerPin):
            if selection == Selection.INSIDE:
                if obj not in found:
                    found.add(obj)
                    yield obj
            else:
                port = obj.port
                if port:
                    definition = port.definition
                    if definition:
                        for instance in definition.references:
                            outer_pin = instance.pins[obj]
                            if outer_pin not in found:
                                found.add(outer_pin)
                                yield outer_pin
        elif isinstance(obj, OuterPin):
            if selection == Selection.OUTSIDE:
                if obj not in found:
                    found.add(obj)
                    yield obj
            else:
                inner_pin = obj.inner_pin
                if inner_pin and inner_pin not in found:
                    found.add(inner_pin)
                    yield inner_pin
        elif isinstance(obj, Wire):
            object_collection += obj.pins
        elif isinstance(obj, Cable):
            object_collection += obj.wires
        elif isinstance(obj, HRef):
            if obj.is_valid:
                item = obj.item
                if isinstance(item, InnerPin) and selection == Selection.OUTSIDE:
                    hport = obj.parent
                    hinstance = hport.parent
                    instance = hinstance.item
                    outer_pin = instance.pins[item]
                    if outer_pin not in found:
                        found.add(outer_pin)
                        yield outer_pin
                else:
                    object_collection.append(obj.item)



from copy import copy, deepcopy, error
import spydrnet as sdn
from spydrnet.ir.element import Element
from spydrnet.ir.outerpin import OuterPin
from spydrnet.ir.views.listview import ListView
from spydrnet.global_state import global_callback


class Wire(Element):
    """
    Represents a wire object
    """
    __slots__ = ['_cable', '_pins', '__weakref__']

    def __init__(self):
        self._cable = None
        self._pins = list()

    @property
    def cable(self):
        """The cable that the wire contains"""
        return self._cable

    @property
    def pins(self):
        """The a list of pins that the wire is connected to"""
        return ListView(self._pins)

    @pins.setter
    def pins(self, value):
        value_list = list(value)
        value_set = set(value_list)
        assert len(value_list) == len(value_set) and set(self._pins) == value_set, \
            "Set of values do not match, assignment can only be used to reorder, values must be unique"
        self._pins = value_list

    def connect_pin(self, pin, position=None):
        """Connects a pin to the wire

        parameters
        ----------

        value - The pin to connect to
        """
        global_callback._call_wire_connect_pin(self, pin)
        if isinstance(pin, OuterPin):
            instance = pin.instance
            inner_pin = pin.inner_pin
            assert instance is not None and inner_pin is not None, \
                "Outer pin must represent an instance and an inner_pin"
            assert inner_pin in instance.pins, "Pin not associated with instance"
            outer_pin = instance.pins[inner_pin]
            assert outer_pin.wire is not self, "Pin already connected to this wire"
            assert outer_pin.wire is None, "Pin already connected to a different wire"
            pin._wire = self
            pin = outer_pin
        else:
            assert pin.wire is None, "Pin already connected to a different wire"
        if position is not None:
            self._pins.insert(position, pin)
        else:
            self._pins.append(pin)
        pin._wire = self

    def disconnect_pin(self, pin):
        """Disconnect a pin from the wire

        parameters
        ----------

        value - The pin to disconnect
        """
        if isinstance(pin, OuterPin):
            instance = pin.instance
            inner_pin = pin.inner_pin
            assert instance is not None and inner_pin is not None, \
                "Outer pin must represent an instance and an inner_pin"
            assert inner_pin in instance.pins, "Pin not associated with instance"
            outer_pin = instance.pins[inner_pin]
            assert outer_pin.wire is self, "Pin is disconnected or connected to a different wire."
            self._disconnect_pin(pin)
            pin = outer_pin
        else:
            assert pin.wire == self, "Pin does not belong to this wire"
        self._pins.remove(pin)
        self._disconnect_pin(pin)

    def disconnect_pins_from(self, pins):
        """Disconnect a list of pins from the wire

        parameters
        ----------

        value - The list of pins to disconnect
        """
        if isinstance(pins, set):
            excluded_pins = pins
        else:
            excluded_pins = set(pins)
        all_pins_can_be_disconnected = True
        for pin in excluded_pins:
            if isinstance(pin, OuterPin):
                instance = pin.instance
                inner_pin = pin.inner_pin
                if instance is None or inner_pin is None or inner_pin not in instance.pins or \
                        instance.pins[inner_pin].wire is not self:
                    all_pins_can_be_disconnected = False
                    break
            else:
                if pin.wire != self:
                    all_pins_can_be_disconnected = False
                    break
        assert all_pins_can_be_disconnected, "Some of the pins to disconnect are not associated with an instance, " \
                                             "already disconnected, or connected to a different wire"
        for pin in excluded_pins:
            if isinstance(pin, OuterPin):
                self._disconnect_pin(pin)
                pin = pin.instance.pins[pin]
            self._disconnect_pin(pin)
        self._pins = list(x for x in self._pins if x not in excluded_pins)

    def _disconnect_pin(self, pin):
        global_callback._call_wire_disconnect_pin(self, pin)
        pin._wire = None

    def _clone_rip_and_replace(self, memo):
        """Remove from its current environment and place it into the new cloned environment with references held in the memo dictionary"""
        new_pins = list()
        for p in self._pins:
            assert p in memo, "the pin must be cloned"
            new_pins.append(memo[p])
        self._pins = new_pins
        pass

    def _clone_rip(self):
        """Remove from its current environmnet.

        This will remove all pin pointers and create a floating stand alone instance."""
        self._pins = list()
        pass

    def _clone(self, memo):
        """Not api safe clone function

        clone leaving all references in tact.
        the element can then either be ripped or ripped and replaced"""
        assert self not in memo, "the object should not have been copied twice in this pass"
        c = Wire()
        memo[self] = c
        c._cable = None
        # shallow copy the list so that it retains its pin references
        c._pins = copy(self._pins)
        return c

    def clone(self):
        """clone wire in an api safe way.

        The following properties can be expected from the returned element:
         * The wire is not connected to any pins.
         * The wire is orphaned from any cable.
         * No pins are connected to the wire
         """
        c = self._clone(dict())
        c._clone_rip()
        return c

    def index(self):
        """if this wire is in a cable, returns the index number of the wire in the parent cable"""

        assert self.cable is not None, "the wire does not belong to a cable"

        return self.cable.wires.index(self)

    def __str__(self):
        """Re-define the print function so it is easier to read"""
        rep = str(type(self))
        rep = rep[:-1] + '; '
        if self.cable is None:
            rep += 'Not contained by any Cable'
        elif self.cable.name is None:
            rep += 'Contained by Cable whose name is undefined'
        else:
            rep += 'Contained by Cable.name \'' + str(self.cable.name) + '\' ' + str(self.cable)
        rep += '>'
        return rep

    def get_driver(self):
        '''
        returns the driver(s) of the wire
        '''
        drivers = []
        for pin in self._pins:
            if pin.__class__ is sdn.InnerPin:
                if pin.port.direction is sdn.IN:
                    drivers.append(pin)
            else:
                if pin.inner_pin.port.direction is sdn.OUT:
                    drivers.append(pin)
        return drivers



import unittest
import spydrnet as sdn


class TestGetWires(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.netlist = sdn.Netlist()

        leaf_library = cls.netlist.create_library()
        leaf_library.name = 'primitives'

        library = cls.netlist.create_library()
        library.name = 'work'

        leaf_def = leaf_library.create_definition()
        leaf_def.name = 'leaf'
        leaf_port = leaf_def.create_port()
        leaf_port.name = 'I'
        leaf_port.create_pins(1)

        bottom_def = library.create_definition()
        bottom_def.name = 'bottom'
        bottom_port = bottom_def.create_port()
        bottom_port.name = 'I'
        bottom_port.create_pins(1)
        leaf_inst = bottom_def.create_child()
        leaf_inst.reference = leaf_def
        bottom_cable = bottom_def.create_cable()
        bottom_cable.name = 'bottom_cable'
        bottom_wire = bottom_cable.create_wire()
        bottom_wire.connect_pin(bottom_port.pins[0])
        bottom_wire.connect_pin(leaf_inst.pins[leaf_port.pins[0]])
        cls.leaf_inst = leaf_inst

        bottom_floating_wire = bottom_cable.create_wire()

        middle_def = library.create_definition()
        middle_def.name = 'middle'
        middle_port = middle_def.create_port()
        middle_port.name = "I"
        middle_port.create_pin()
        bottom_inst = middle_def.create_child()
        bottom_inst.name = 'bottom'
        bottom_inst.reference = bottom_def
        middle_cable = middle_def.create_cable()
        middle_cable.name = "middle_cable"
        middle_wire = middle_cable.create_wire()
        middle_wire.connect_pin(middle_port.pins[0])
        middle_wire.connect_pin(bottom_inst.pins[bottom_port.pins[0]])
        cls.bottom_inst = bottom_inst

        middle_floating_wire = middle_cable.create_wire()

        top_def = library.create_definition()
        top_def.name = 'top'
        top_port = top_def.create_port()
        top_port.name = "I"
        top_port.create_pin()
        middle_inst = top_def.create_child()
        middle_inst.name = 'middle'
        middle_inst.reference = middle_def
        top_cable = top_def.create_cable()
        top_cable.name = "top_cable"
        top_wire = top_cable.create_wire()
        top_wire.connect_pin(top_port.pins[0])
        top_wire.connect_pin(middle_inst.pins[middle_port.pins[0]])
        cls.middle_inst = middle_inst

        top_floating_wire = top_cable.create_wire()

        top_instance = sdn.Instance()
        top_instance.name = 'top'
        top_instance.reference = top_def
        cls.netlist.top_instance = top_instance

    def test_parameter_checking(self):
        library = sdn.Library()
        definition = library.create_definition()
        definition.name = "MY_DEF"
        self.assertRaises(TypeError, sdn.get_wires, library, "MY_DEF")
        self.assertRaises(TypeError, sdn.get_wires, library, unsupported_keyword=None)
        self.assertRaises(TypeError, sdn.get_wires, library, selection=None)
        self.assertRaises(TypeError, sdn.get_wires, None)
        self.assertRaises(TypeError, sdn.get_wires, [None, library])

    def test_collection(self):
        wires = list(sdn.get_wires([self.netlist.libraries[1], self.netlist.libraries[1]]))
        self.assertEqual(len(wires), 6)

    def test_inside(self):
        wires = list(self.middle_inst.get_wires(selection="INSIDE", recursive=True))
        self.assertEqual(len(wires), 4)

    def test_both(self):
        wires = list(sdn.get_wires(self.middle_inst, selection="BOTH", recursive=True))
        self.assertEqual(len(wires), 2)

    def test_outside(self):
        wires = list(sdn.get_wires(self.middle_inst, selection="OUTSIDE", recursive=True))
        self.assertEqual(len(wires), 1)

    def test_wire_inside(self):
        wires = list(sdn.get_wires(self.middle_inst.reference.cables[0].wires[0]))
        self.assertEqual(len(wires), 1)
        self.assertIs(wires[0], self.middle_inst.reference.cables[0].wires[0])

    def test_wire_outside(self):
        wires = list(sdn.get_wires(self.middle_inst.reference.cables[0].wires[0], selection="OUTSIDE"))
        self.assertEqual(len(wires), 2)
        self.assertTrue(self.middle_inst.reference.cables[0].wires[0] not in wires)

    def test_port_outside(self):
        wires = list(sdn.get_wires(self.middle_inst.reference.ports[0], selection="OUTSIDE"))
        self.assertEqual(len(wires), 1)
        self.assertTrue(wires[0].cable.name == 'top_cable')

    def test_cable_outside(self):
        wires = list(sdn.get_wires(self.middle_inst.reference.cables[0], selection="OUTSIDE"))
        self.assertEqual(len(wires), 2)
        self.assertTrue(all(x not in wires for x in self.middle_inst.reference.cables[0].wires))

    def test_cable_all(self):
        wires = list(sdn.get_wires(self.netlist.top_instance.reference.cables[0].wires[0], selection="ALL"))
        self.assertEqual(len(wires), 3)

    def test_pin_all(self):
        wires = list(sdn.get_wires(self.middle_inst.reference.ports[0].pins[0], selection="ALL"))
        self.assertEqual(len(wires), 3)

    def test_pin_inside(self):
        wires = list(sdn.get_wires(self.middle_inst.reference.ports[0].pins[0]))
        self.assertEqual(len(wires), 1)
        self.assertIs(wires[0], self.middle_inst.reference.cables[0].wires[0])

    def test_outer_pin_inside(self):
        inner_pin = self.middle_inst.reference.ports[0].pins[0]
        outer_pin = self.middle_inst.pins[inner_pin]
        wires = list(sdn.get_wires(outer_pin))
        self.assertEqual(len(wires), 1)
        self.assertIs(wires[0], self.middle_inst.reference.cables[0].wires[0])

    def test_outer_pin_outside(self):
        inner_pin = self.middle_inst.reference.ports[0].pins[0]
        outer_pin = self.middle_inst.pins[inner_pin]
        wires = list(sdn.get_wires(outer_pin, selection="OUTSIDE"))
        self.assertEqual(len(wires), 1)
        self.assertIs(wires[0], self.netlist.top_instance.reference.cables[0].wires[0])

    def test_netlist(self):
        wires = list(sdn.get_wires(self.netlist))
        self.assertEqual(len(wires), 6)

    def test_href(self):
        href = next(sdn.get_hinstances(self.netlist.top_instance))
        wires = list(sdn.get_wires(href, recursive=True))
        self.assertEqual(len(wires), 6)



import unittest
import spydrnet as sdn
from spydrnet.util.hierarchical_reference import HRef


class TestHRefBase(unittest.TestCase):
    def test_is_unique(self):
        netlist = sdn.Netlist()
        library = netlist.create_library()
        definition1 = library.create_definition()
        instance1 = definition1.create_child()

        definition2 = library.create_definition()
        instance2a = definition2.create_child()
        instance2a.reference = definition1
        instance2b = definition2.create_child()
        instance2b.reference = definition1

        instance3 = sdn.Instance()
        instance3.reference = definition2
        netlist.top_instance = instance3

        href1 = HRef.from_sequence([instance3, instance2a, instance1])
        href2 = HRef.from_sequence([instance3, instance2b, instance1])

        # Another test for test_flyweight stuffed in here :)
        self.assertTrue(href1.parent.parent is href2.parent.parent)
        self.assertTrue(href1.is_valid)
        self.assertTrue(href2.is_valid)
        self.assertFalse(href2.is_unique)

        self.assertFalse(href1.is_unique)
        definition2.remove_child(instance2b)
        self.assertFalse(href2.is_valid)
        self.assertFalse(href2.is_unique)
        self.assertTrue(href1.is_unique)

        definition4 = library.create_definition()
        definition4.add_child(instance2b)
        instance4 = sdn.Instance()
        instance4.reference = definition4
        definition2.add_child(instance4)
        self.assertFalse(href1.is_unique)

    def test_flyweight(self):
        instance = sdn.Instance()
        href1 = HRef.from_parent_and_item(None, instance)
        href2 = HRef.from_parent_and_item(None, instance)
        self.assertTrue(href1 is href2)
        import weakref
        w_href1 = weakref.ref(href1)
        w_href2 = weakref.ref(href2)
        href1 = None
        href2 = None
        import gc
        gc.collect()
        self.assertIsNone(w_href1())
        self.assertIsNone(w_href2())

        instance2 = sdn.Instance()
        href1 = HRef.from_sequence([instance, instance2])
        href2_parent = HRef(instance, None)
        href2 = HRef.from_parent_and_item(href2_parent, instance2)
        self.assertTrue(href1 is href2)

    def test_href_valid(self):
        instance = sdn.Instance()
        href = HRef.from_parent_and_item(None, instance)
        self.assertFalse(href.is_valid)

        definition = sdn.Definition()
        instance.reference = definition
        self.assertFalse(href.is_valid)

        library = sdn.Library()
        library.add_definition(definition)
        self.assertFalse(href.is_valid)

        netlist = sdn.Netlist()
        netlist.add_library(library)
        self.assertFalse(href.is_valid)

        netlist.top_instance = instance
        self.assertTrue(href.is_valid)

        cable = sdn.Cable()
        wire = sdn.Wire()

        href = HRef.from_sequence([instance, cable, wire])
        self.assertFalse(href.is_valid)

        cable.add_wire(wire)
        self.assertFalse(href.is_valid)

        definition.add_cable(cable)
        self.assertTrue(href.is_valid)

        instance.reference = None
        self.assertFalse(href.is_valid)

        port = sdn.Port()
        pin  = sdn.InnerPin()

        href = HRef.from_sequence([instance, port, pin])
        self.assertFalse(href.is_valid)

        port.add_pin(pin)
        self.assertFalse(href.is_valid)

        definition.add_port(port)
        self.assertFalse(href.is_valid)

        instance.reference = definition
        self.assertTrue(href.is_valid)

        higher_definition = library.create_definition()
        higher_definition.add_child(instance)
        self.assertTrue(href.is_valid)

        higher_instance = sdn.Instance()
        higher_instance.reference = higher_definition
        netlist.top_instance = higher_instance
        self.assertFalse(href.is_valid)

        href = HRef.from_sequence([higher_instance, instance, cable, wire])
        self.assertTrue(href.is_valid)
        higher_instance.reference = None
        self.assertFalse(href.is_valid)
        higher_definition.remove_child(instance)
        self.assertFalse(href.is_valid)

        definition.remove_cable(cable)
        self.assertFalse(href.is_valid)

        cable.remove_wire(wire)
        self.assertFalse(href.is_valid)

        cable.add_wire(wire)
        href = HRef.from_sequence([cable, wire])
        self.assertFalse(href.is_valid)

        cable.remove_wire(wire)
        new_cable = sdn.Cable()
        new_cable.add_wire(wire)
        self.assertFalse(href.is_valid)

        href = HRef.from_parent_and_item(None, wire)
        self.assertFalse(href.is_valid)

        port.remove_pin(pin)
        new_port = sdn.Port()
        new_port.add_pin(pin)
        href = HRef.from_sequence([port, pin])
        self.assertFalse(href.is_valid)

        href = HRef.from_parent_and_item(None, pin)
        self.assertFalse(href.is_valid)

        href = HRef.from_parent_and_item(None, definition)
        self.assertFalse(href.is_valid)

    def test_href_equality(self):
        top = sdn.Instance()
        top_def = sdn.Definition()
        top.reference = top_def
        middle = top_def.create_child()
        middle_def = sdn.Definition()
        middle.reference = middle_def
        leaf = middle_def.create_child()
        leaf_def = sdn.Definition()
        leaf.reference = leaf_def
        sequence = [top, middle, leaf]
        href1 = HRef.from_sequence(sequence)
        href2 = HRef.from_sequence(sequence)
        self.assertEqual(href1, href1)
        self.assertNotEqual(href1, None)
        self.assertEqual(href1, href2)
        self.assertNotEqual(href1, href2.parent)
        href3 = HRef.from_sequence(sequence[1:])
        self.assertNotEqual(href1, href3)

    def test_href_inst_name(self):
        instance = sdn.Instance()
        instance.name = "MY_INST"
        href = HRef.from_parent_and_item(None, instance)
        self.assertEqual('', href.name)

    def test_href_wire_name(self):
        top = sdn.Instance()
        top.name = "TOP"
        top_def = sdn.Definition()
        top.reference = top_def
        cable = top_def.create_cable()
        cable.name = "CABLE"
        cable.create_wires(1)
        cable.is_array = True
        wire = cable.wires[0]
        sequence = [top, cable, wire]
        self.assertEqual("CABLE[0]", HRef.from_sequence(sequence).name)

    def test_href_pin_name(self):
        top = sdn.Instance()
        top.name = "TOP"
        top_def = sdn.Definition()
        top.reference = top_def
        port = top_def.create_port()
        port.name = "PORT"
        port.create_pins(1)
        port.is_array = True
        wire = port.pins[0]
        sequence = [top, port, wire]
        self.assertEqual("PORT[0]", HRef.from_sequence(sequence).name)

    def test_href_str_and_repr(self):
        netlist = sdn.Netlist()
        library = netlist.create_library()
        definition = library.create_definition()
        instance = sdn.Instance()
        instance.reference = definition
        netlist.top_instance = instance

        cable = definition.create_cable()
        cable.name = "MY_CABLE"
        cable.create_wires(8)
        href = HRef.from_sequence([instance, cable])
        href_str = str(href)
        self.assertTrue(href_str == "MY_CABLE")
        href_repr = repr(href)
        self.assertTrue(HRef.__name__ in href_repr)
        self.assertTrue(cable.__class__.__name__ in href_repr)

    def test_get_all_hrefs_of_item(self):
        netlist = sdn.Netlist()
        library = netlist.create_library()
        definition = library.create_definition()
        instance = sdn.Instance()
        instance.reference = definition
        netlist.top_instance = instance

        href = next(HRef.get_all_hrefs_of_item(instance))
        self.assertTrue(href.item is instance)

        href = next(HRef.get_all_hrefs_of_item(definition))
        self.assertTrue(href.item is instance)

        port = definition.create_port()
        href = next(HRef.get_all_hrefs_of_item(port))
        self.assertTrue(href.is_valid and href.item is port)



from collections import deque, OrderedDict
from spydrnet.ir.port import Port
from spydrnet.ir.cable import Cable
import spydrnet.parsers.verilog.verilog_tokens as vt


class Composer:

    def __init__(self, definition_list=None, write_blackbox=False):
        """ Write a verilog netlist from SDN netlist

        parameters
        ----------

        definition_list - (list[str]) list of defintions to write
        write_blackbox - (bool) Skips writing black boxes/verilog primitives
        """
        self.file = None
        self.direction_string_map = dict()
        self.direction_string_map[Port.Direction.IN] = "input"
        self.direction_string_map[Port.Direction.OUT] = "output"
        self.direction_string_map[Port.Direction.INOUT] = "inout"
        self.direction_string_map[Port.Direction.UNDEFINED] = "/* undefined port direction */ inout"
        self.written = set()
        self.indent_count = 4  # set the indentation level for various components
        self.write_blackbox = write_blackbox
        self.definition_list = definition_list


    def run(self, ir, file_out="out.v"):
        self._open_file(file_out)
        self._compose(ir)

    def _open_file(self, file_name):
        f = open(file_name, "w")
        self.file = f

    def _compose(self, netlist):
        self._write_header(netlist)
        instance = netlist.top_instance
        if instance is not None:
            self._write_from_top(instance)
        for library in netlist.libraries:
            for definition in library.definitions:
                if definition not in self.written:
                    self._write_module(definition)

    def _write_header(self, netlist):
        self.file.write("//Generated from netlist by SpyDrNet\n")
        self.file.write("//netlist name: " + netlist.name + "\n")

    def _write_from_top(self, instance):
        #self.written = set()
        to_write = deque()
        to_write.append(instance.reference)
        while(len(to_write) != 0):
            definition = to_write.popleft()
            if definition in self.written:
                continue
            self.written.add(definition)
            for c in definition.children:
                if c.reference not in self.written:
                    to_write.append(c.reference)
            assert definition.name is not None, self._error_string(
                "definition has no name set", definition)
            self._write_module(definition)


    ###########################################################################
    # Write verilog constructs
    ###########################################################################

    def _write_preprocessor_directives(self, module):
        '''write out the backtick preprocessor directives.
        this will likely just be `timescale 1 ps / 1 ps

        UPDATE: I'm just going to skip this for now.
        It can be added if simulation support is needed later'''
        pass  # not currently implemented, just pass

    def _write_star_constraints(self, o):
        if "VERILOG.InlineConstraints" in o and len(o["VERILOG.InlineConstraints"]) != 0:
            dictionary = o["VERILOG.InlineConstraints"]
            self.file.write(vt.OPEN_PARENTHESIS)
            self.file.write(vt.STAR)
            self.file.write(vt.SPACE)
            first = ""
            for k, v in dictionary.items():
                if v is not None:
                    self.file.write(first + k + vt.SPACE +
                                    vt.EQUAL + vt.SPACE + v)
                else:
                    self.file.write(first + k)
                first = vt.COMMA + vt.SPACE
            self.file.write(vt.SPACE)
            self.file.write(vt.STAR)
            self.file.write(vt.CLOSE_PARENTHESIS)
            self.file.write(vt.NEW_LINE)

    def _write_bundle_with_indicies(self, bundle, low, high):
        '''write out a bundle name and indicies. in name indicies order.
        useful for cable and port instances'''
        self._write_name(bundle)
        self._write_brackets(bundle, low, high)

    def _write_module(self, definition):
        '''write the constraints then the module header then the module body'''
        if self.definition_list:
            if not (definition.name in self.definition_list):
                return
        if definition.library.name == "SDN_VERILOG_ASSIGNMENT":
            return  # don't write assignment definitions
        if definition.library.name == "SDN.verilog_primitives":
            if not self.write_blackbox:
                return
            self.file.write(vt.CELL_DEFINE)
            self.file.write(vt.NEW_LINE)
        self._write_star_constraints(definition)
        self._write_module_header(definition)
        self._write_module_body(definition)
        self.file.write(vt.END_MODULE)
        if definition.library.name == "SDN.verilog_primitives":
            self.file.write(vt.NEW_LINE)
            self.file.write(vt.END_CELL_DEFINE)

        self.file.write(2 * vt.NEW_LINE)

    def _write_module_header(self, definition):
        '''write out the module header with the following style:
        module module_name (port_name, port_name);
        start with module and end with semi colon
        '''
        self.file.write(vt.MODULE)
        self.file.write(vt.SPACE)
        self._write_name(definition)
        self.file.write(vt.NEW_LINE)
        if "VERILOG.Parameters" in definition:
            self._write_module_parameters(definition)
        self._write_module_header_ports(definition)
        self.file.write(vt.NEW_LINE)

    def _write_module_body(self, definition):
        '''write out the module body start with ports, then do assignments then instances'''
        self._write_module_body_ports(definition)
        if definition.library.name != "SDN.verilog_primitives":
            self._write_module_body_cables(definition)
            self._write_assignments(definition)
            self._write_module_body_instances(definition)

    def _write_module_body_instances(self, definition):
        for c in definition.children:
            self._write_module_body_instance(c)

    def _write_module_body_instance(self, instance):
        if instance.reference.library.name == "SDN_VERILOG_ASSIGNMENT":
            return  # do not write the assignment instance
        self._write_star_constraints(instance)
        self.file.write(self.indent_count * vt.SPACE)
        self._write_name(instance.reference)
        self.file.write(vt.SPACE)
        if "VERILOG.Parameters" in instance:
            self._write_instance_parameters(instance)
            self.file.write(self.indent_count * vt.SPACE)
        self._write_name(instance)
        self.file.write(vt.NEW_LINE)
        self._write_instance_ports(instance)
        self.file.write(vt.NEW_LINE)

    def _write_module_body_ports(self, definition):
        for p in definition.ports:
            self._write_module_body_port(p)
        self.file.write(vt.NEW_LINE)

    def _write_module_body_port(self, port):
        _, cables = self._all_wires_and_cables_from_pinset(port.pins)
        if len(cables) == 0:
            cables.append(port) #adding the port will let composer to still print out disconnected ports
        for c in cables:
            self._write_star_constraints(port)
            self.file.write(self.indent_count * vt.SPACE)
            self.file.write(self.direction_string_map[port.direction])
            self.file.write(vt.SPACE)
            self._write_brackets_defining(c)
            self._write_name(c)
            self.file.write(vt.SEMI_COLON)
            self.file.write(vt.NEW_LINE)

    def _write_module_body_cables(self, definition):
        for c in definition.cables:
            self._write_module_body_cable(c)
        self.file.write(vt.NEW_LINE)

    def _write_module_body_cable(self, cable):
        self._write_star_constraints(cable)
        self.file.write(self.indent_count * vt.SPACE)
        if "VERILOG.CableType" in cable:
            self.file.write(cable["VERILOG.CableType"])
        else:
            self.file.write(vt.WIRE)
        self.file.write(vt.SPACE)
        self._write_brackets_defining(cable)
        self._write_name(cable)
        self.file.write(vt.SEMI_COLON)
        self.file.write(vt.NEW_LINE)

    def _write_concatenation(self, wires):
        '''write out the concatenation statement
        {wire_1, wire_2}'''
        self.file.write(vt.OPEN_BRACE)
        first = True
        for w in wires:
            if w is not None:
                if not first:
                    self.file.write(vt.COMMA)
                    self.file.write(vt.SPACE)
                index = self._index_of_wire_in_cable(w)
                self._write_bundle_with_indicies(w.cable, index, None)
            else:
                break

            first = False
        self.file.write(vt.CLOSE_BRACE)

    def _write_module_header_ports(self, definition):
        self.file.write(vt.OPEN_PARENTHESIS)
        first = True
        for p in definition.ports:
            if not first:
                self.file.write(vt.COMMA)
            self.file.write(vt.NEW_LINE)
            self._write_module_header_port(p)
            first = False
        self.file.write(vt.NEW_LINE)
        self.file.write(vt.CLOSE_PARENTHESIS)
        self.file.write(vt.SEMI_COLON)
        self.file.write(vt.NEW_LINE)

    def _write_module_header_port(self, port):
        '''does not write the input output or width, 
        check for concatenation port as well'''
        self.file.write((self.indent_count * vt.SPACE))
        aliased = self._is_pinset_concatenated(port.pins, port.name)
        if aliased:
            wires = []
            for p in port.pins:
                wires.append(p.wire)
            self.file.write(vt.DOT)
            self._write_name(port)
            self.file.write(vt.OPEN_PARENTHESIS)
            self._write_concatenation(wires)
            self.file.write(vt.CLOSE_PARENTHESIS)
        else:
            self._write_name(port)

    def _write_assignments(self, definition):
        for c in definition.children:
            if c.reference.library.name == "SDN_VERILOG_ASSIGNMENT":
                self._write_assignment(c)

    def _write_assignment(self, instance):
        '''take an assignment instance and write an assignment out
        assign cable_out = cable_in;'''
        in_port = None
        out_port = None
        for p in instance.reference.ports:
            if p.name == "o":
                out_port = p
            if p.name == "i":
                in_port = p
        assert in_port is not None, self._error_string(
            "instance does not appear to be an assignment with a port named i and o", instance)
        assert out_port is not None, self._error_string(
            "instance does not appear to be an assignment with a port named i and o", instance)
        left_wires = []
        right_wires = []
        in_pins = []
        out_pins = []
        for p in in_port.pins:
            in_pins.append(instance.pins[p])
        for p in out_port.pins:
            out_pins.append(instance.pins[p])
        in_wires, in_cables = self._all_wires_and_cables_from_pinset(in_pins)
        out_wires, out_cables = self._all_wires_and_cables_from_pinset(out_pins)
        assert not self._is_pinset_concatenated(in_pins, in_wires[0].cable.name), self._error_string(
            "multiple cables appear to be connected to a single assignment input", instance)
        assert not self._is_pinset_concatenated(out_pins, out_wires[0].cable.name), self._error_string(
            "multiple cables appear to be connected to a single assignment output", instance)
        self.file.write(vt.ASSIGN)
        self.file.write(vt.SPACE)
        hi = self._index_of_wire_in_cable(out_wires[-1])
        li = self._index_of_wire_in_cable(out_wires[0])
        self._write_bundle_with_indicies(out_cables[0], li, hi)
        self.file.write(vt.SPACE)
        self.file.write(vt.EQUAL)
        self.file.write(vt.SPACE)
        hi = self._index_of_wire_in_cable(in_wires[-1])
        li = self._index_of_wire_in_cable(in_wires[0])
        self._write_bundle_with_indicies(in_cables[0], li, hi)
        self.file.write(vt.SEMI_COLON)
        self.file.write(vt.NEW_LINE)

    def _write_module_parameters(self, definition):
        '''write out the parameters in the module header'''
        self.file.write(vt.OCTOTHORP)
        self.file.write(vt.OPEN_PARENTHESIS)
        indentation = "\n    "
        first = True
        for k, v in definition["VERILOG.Parameters"].items():
            if not first:
                self.file.write(vt.COMMA)
            self.file.write(indentation)
            self.file.write(vt.PARAMETER)
            self.file.write(vt.SPACE)
            self.file.write(k)
            if v is not None:
                self.file.write(" " + vt.EQUAL + " ")
                self.file.write(v)
            first = False
        self.file.write(vt.NEW_LINE)
        self.file.write(vt.CLOSE_PARENTHESIS)

    def _write_instance_parameters(self, instance):
        '''write out the parameters in the instance
        #(
        .key(value),
        ...
        )'''
        self.file.write(vt.OCTOTHORP)
        self.file.write(vt.OPEN_PARENTHESIS)
        self.file.write(vt.NEW_LINE)
        first = True
        for k, v in instance["VERILOG.Parameters"].items():
            if not first:
                self.file.write(vt.COMMA)
                self.file.write(vt.NEW_LINE)
            self._write_instance_parameter(k, v)
            first = False
        self.file.write(vt.NEW_LINE)
        self.file.write(self.indent_count * vt.SPACE)
        self.file.write(vt.CLOSE_PARENTHESIS)
        self.file.write(vt.NEW_LINE)

    def _write_instance_parameter(self, key, value):
        self.file.write(2*self.indent_count * vt.SPACE)
        self.file.write(vt.DOT)
        self.file.write(key)
        self.file.write(vt.OPEN_PARENTHESIS)
        self.file.write(value)
        self.file.write(vt.CLOSE_PARENTHESIS)

    def _write_instance_ports(self, instance):
        '''write out the port mapping on an instance.'''
        self.file.write(self.indent_count * vt.SPACE)
        self.file.write(vt.OPEN_PARENTHESIS)
        self.file.write(vt.NEW_LINE)
        first = True
        for p in instance.reference.ports:
            if not first:
                self.file.write(vt.COMMA)
                self.file.write(vt.NEW_LINE)
            self._write_instance_port(instance, p)
            first = False
        self.file.write(vt.NEW_LINE)
        self.file.write(self.indent_count * vt.SPACE)
        self.file.write(vt.CLOSE_PARENTHESIS)
        self.file.write(vt.SEMI_COLON)

    def _write_instance_port(self, instance, port):
        self.file.write(2*self.indent_count * vt.SPACE)
        self.file.write(vt.DOT)
        self._write_name(port)
        self.file.write(vt.OPEN_PARENTHESIS)
        pins = []
        for p in port.pins:
            pins.append(instance.pins[p])
        if pins[0].wire is not None:
            name = pins[0].wire.cable.name
        else:
            name = None
        concatenated = self._is_pinset_concatenated(pins, name)
        wires = []
        for p in pins:
            wires.append(p.wire)
        if concatenated:
            self._write_concatenation(wires)
        else:
            if pins[0].wire is not None:
                last = -1
                wl = wires[last]
                wr = wires[0]
                while wl is None:  # get the last named non none wire.
                    last = last - 1
                    wl = wires[last]
                il = self._index_of_wire_in_cable(wl)
                ir = self._index_of_wire_in_cable(wr)
                self._write_bundle_with_indicies(wl.cable, ir, il)

        self.file.write(vt.CLOSE_PARENTHESIS)

    def _write_name(self, o):
        '''write the name of an o. this is split out to give an error message if the name is not set
        In the future this could be changed to add a name to os that do not have a name set'''
        assert o.name is not None, self._error_string("name of o is not set", o)
        if o.name[0] == '\\':
            assert o.name[-1] == ' ', self._error_string(
                "the o name starts with escape and does not end with a space.", o)
        self.file.write(o.name)

    def _write_brackets_defining(self, bundle):
        '''write the brackets for port or cable definitions'''
        if isinstance(bundle, Cable):
            array = bundle.wires
        elif isinstance(bundle, Port):
            array = bundle.pins
        width = len(array)

        assert width != 0, self._error_string(
            "bundle has 0 width, this will not write correctly please add a pin to the port or wire to the cable.", bundle)

        if width == 1 and bundle.lower_index == 0:
            return  # no need to write because this is assumed

        self.file.write(vt.OPEN_BRACKET)
        self.file.write(str(bundle.lower_index + width - 1))
        self.file.write(vt.COLON)
        self.file.write(str(bundle.lower_index))
        self.file.write(vt.CLOSE_BRACKET)

    def _write_brackets(self, bundle, low_index, high_index):
        '''write a bundle's brackets based on the low and high indicies given
        does not write out the name, works on both cables and ports'''
        if low_index == None and high_index == None:
            return  # nothing to write.

        if isinstance(bundle, Cable):
            array = bundle.wires
        elif isinstance(bundle, Port):
            array = bundle.pins
        width = len(array)
        assert width != 0, self._error_string(
            "cannot index into 0 width cable")
        lower_bundle = bundle.lower_index
        upper_bundle = lower_bundle + width - 1

        #intended logic
        #the bundle is a single bit: assert indicies are within but nothing to be written
        #the bundle is multibit and the indicies match the upper and lower(or none): nothing to be written
        #the bundle is multibit but the indicies match each other or one is none: write a single index
        #the bundle is multibit but the indicies don't match each other: write both indicies
        
        if width == 1:
            assert (low_index == None or low_index == lower_bundle), \
                self._error_string("attempted to index bundle out of bounds at " + str(low_index), bundle)
            assert (high_index == None or high_index == upper_bundle), \
                self._error_string("attempted to index bundle out of bounds at " + str(high_index), bundle)
            return
        elif (low_index == lower_bundle and high_index == upper_bundle) or (low_index == None and high_index == None):
            return
        elif low_index == high_index or low_index is None or high_index is None:
            index = low_index
            if index is None:
                index = high_index
            #this assertion checks the logic of this if statement
            assert index is not None, self._error_string("if both high and low indicies are None no brackets need to be written", bundle)
            #the following assertion will check the inputs to the function
            assert index >= lower_bundle and index <= upper_bundle,\
                self._error_string("attempted to write an index out of bounds: " + str(index), bundle)
            self.file.write("[" + str(index) + "]")
        else:
            #first assertion is an internal error, check the if elif logic here.
            assert low_index != high_index and low_index is not None and high_index is not None,\
                self._error_string("a single value needs to be written if any of these conditions are true ", bundle)
            #these assertions highlight issues with the input to the function
            assert low_index >= lower_bundle and low_index <= upper_bundle,\
                self._error_string("attempted to write an index out of bounds: " + str(low_index), bundle)
            assert high_index >= lower_bundle and high_index <= upper_bundle,\
                self._error_string("attempted to write an index out of bounds: " + str(high_index), bundle)
            self.file.write("[" + str(high_index) + ":" + str(low_index) + "]")


    ###############################################################################
    # helper functions for composing
    ###############################################################################

    def _index_of_wire_in_cable(self, wire):
        index = 0
        for w in wire.cable.wires:
            if w == wire:
                return index + wire.cable.lower_index
            index += 1
        return None

    def _is_pinset_concatenated(self, pins, name):
        aliased = False
        now_none = False
        last_index = None
        for p in pins:
            if p.wire is None:
                next_name = None
                now_none = True
            else:
                next_name = p.wire.cable.name
                if now_none == True:
                    aliased = True
                index = self._index_of_wire_in_cable(p.wire)
                if last_index == None:
                    last_index = index
                else:
                    if index != last_index + 1:
                        aliased = True
                        break
                last_index = index
            if next_name != name and not now_none:
                aliased = True
                break
            
        return aliased

    def _all_wires_and_cables_from_pinset(self, pins):
        wires = []
        cables = []
        for p in pins:
            wires.append(p.wire)
            if p.wire is not None and p.wire.cable not in cables:
                cables.append(p.wire.cable)
        return wires, cables

    def _error_string(self, message, o):
        return "compose error: " + message + str(o)

    def _write_escapable_name(self, str_in):
        if str_in[0] == "\\":
            self.file.write(str_in + " ")
        else:
            self.file.write(str_in)



import unittest
import spydrnet as sdn


class TestGetCables(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.netlist = sdn.Netlist()

        leaf_library = cls.netlist.create_library()
        leaf_library.name = 'primitives'

        library = cls.netlist.create_library()
        library.name = 'work'

        leaf_def = leaf_library.create_definition()
        leaf_def.name = 'leaf'
        leaf_port = leaf_def.create_port()
        leaf_port.name = 'I'
        leaf_port.create_pins(1)

        bottom_def = library.create_definition()
        bottom_def.name = 'bottom'
        bottom_port = bottom_def.create_port()
        bottom_port.name = 'I'
        bottom_port.create_pins(1)
        leaf_inst = bottom_def.create_child()
        leaf_inst.reference = leaf_def
        bottom_cable = bottom_def.create_cable()
        bottom_cable.name = 'bottom_cable'
        bottom_wire = bottom_cable.create_wire()
        bottom_wire.connect_pin(bottom_port.pins[0])
        bottom_wire.connect_pin(leaf_inst.pins[leaf_port.pins[0]])
        cls.leaf_inst = leaf_inst

        bottom_floating_wire = bottom_cable.create_wire()

        middle_def = library.create_definition()
        middle_def.name = 'middle'
        middle_port = middle_def.create_port()
        middle_port.name = "I"
        middle_port.create_pin()
        bottom_inst = middle_def.create_child()
        bottom_inst.name = 'bottom'
        bottom_inst.reference = bottom_def
        middle_cable = middle_def.create_cable()
        middle_cable.name = "middle_cable"
        middle_wire = middle_cable.create_wire()
        middle_wire.connect_pin(middle_port.pins[0])
        middle_wire.connect_pin(bottom_inst.pins[bottom_port.pins[0]])
        cls.bottom_inst = bottom_inst

        middle_floating_wire = middle_cable.create_wire()

        top_def = library.create_definition()
        top_def.name = 'top'
        top_port = top_def.create_port()
        top_port.name = "I"
        top_port.create_pin()
        middle_inst = top_def.create_child()
        middle_inst.name = 'middle'
        middle_inst.reference = middle_def
        top_cable = top_def.create_cable()
        top_cable.name = "top_cable"
        top_wire = top_cable.create_wire()
        top_wire.connect_pin(top_port.pins[0])
        top_wire.connect_pin(middle_inst.pins[middle_port.pins[0]])
        cls.middle_inst = middle_inst

        top_floating_wire = top_cable.create_wire()

        top_instance = sdn.Instance()
        top_instance.name = 'top'
        top_instance.reference = top_def
        cls.netlist.top_instance = top_instance

    def test_parameter_checking(self):
        definition = sdn.Definition()
        cable = definition.create_cable()
        cable.name = "MY_CABLE"
        self.assertRaises(TypeError, sdn.get_cables, definition, "MY_CABLE", patterns="MY_CABLE")
        self.assertRaises(TypeError, sdn.get_cables, definition, "MY_CABLE", unsupported_keyword=None)
        self.assertRaises(TypeError, sdn.get_cables, definition, "MY_CABLE", selection=None)
        self.assertRaises(TypeError, sdn.get_cables, None, "MY_CABLE")
        self.assertRaises(TypeError, sdn.get_cables, [None, definition], "MY_CABLE")

    def test_collection(self):
        definition = sdn.Definition()
        cable = definition.create_cable()
        cable.name = "MY_CABLE"
        cables = list(sdn.get_cables([definition, definition]))
        self.assertEqual(len(cables), 1)

    def test_get_cables_in_library(self):
        library = sdn.Library()
        definition = library.create_definition()
        cable = definition.create_cable()
        cable.name = "MY_PORT"
        instance = sdn.Instance()
        instance.reference = definition
        port1 = next(library.get_cables("MY_PORT"))
        self.assertEqual(cable, port1)

    def test_get_cables_in_netlist(self):
        netlist = sdn.Netlist()
        library = netlist.create_library()
        definition = library.create_definition()
        cables = definition.create_cable()
        cables.name = "MY_PORT"
        instance = sdn.Instance()
        instance.reference = definition
        port1 = next(netlist.get_cables("MY_PORT"))
        self.assertEqual(cables, port1)

    def test_select_all(self):
        middle_def = self.middle_inst.reference
        search = list(sdn.get_cables(middle_def, selection="ALL"))
        self.assertEqual(len(search), 3)

    def test_href_to_cable(self):
        href = next(sdn.get_hcables(self.middle_inst))
        search = next(sdn.get_cables(href))
        self.assertEqual(href.item, search)

    def test_href_to_cable_all(self):
        href = next(sdn.get_hcables(self.middle_inst))
        search = list(sdn.get_cables(href, selection="ALL"))
        self.assertEqual(len(search), 3)

    def test_port(self):
        port = self.netlist.top_instance.reference.ports[0]
        search = list(sdn.get_cables(port))
        self.assertEqual(len(search), 1)
        self.assertEqual(port.pins[0].wire.cable, search[0])

    def test_outer_pin(self):
        search = list(sdn.get_cables(self.middle_inst.pins, selection="ALL"))
        self.assertEqual(len(search), 3)

    def test_wire_inside(self):
        search = list(sdn.get_cables(self.netlist.top_instance.reference.cables[0].wires[0]))
        self.assertEqual(len(search), 1)
        self.assertEqual(search[0].name, "top_cable")

    def test_wire_outside(self):
        search = list(sdn.get_cables(self.netlist.top_instance.reference.cables[0].wires[0], selection="OUTSIDE"))
        self.assertEqual(len(search), 1)
        self.assertEqual(search[0].name, "middle_cable")

    def test_wires_all(self):
        search = list(sdn.get_cables(self.netlist.top_instance.reference.cables[0].wires[0], selection="ALL"))
        self.assertEqual(len(search), 3)

    def test_absolute_name_from_relative_reference(self):
        search = next(sdn.get_cables(self.middle_inst, "middle_cable"))
        self.assertEqual(search, self.middle_inst.reference.cables[0])



# Copyright 2020 Dallin Skouson, Andrew Keller, Michael Wirthlin

from spydrnet.ir import *
from collections import deque
from spydrnet.uniquify import uniquify

"""How to flatten (brainstorm):
start at the top and take all of the subelements out and add them to the top definition
Naming will be dificult but I will just skip that for now to get it working.
"""

unique_number = 0

mod_name_uid = 0


def _get_unique_name_modifier():
    global mod_name_uid
    str_out = "sdn_flat_" + str(mod_name_uid)
    mod_name_uid += 1
    return str_out


def _rename_element(e):
    global unique_number
    e.name = e.name + "_flat_" + str(unique_number)
    unique_number += 1


def _redo_connections(instance, port):
    for pin in port.pins:
        out_pin = instance.pins[pin]
        out_wire = out_pin.wire
        in_pin = pin
        in_wire = in_pin.wire

        if in_wire:
            in_wire.disconnect_pin(in_pin)
        if out_wire:
            out_wire.disconnect_pin(out_pin)

        pins_to_move = []
        if in_wire:
            for p in in_wire.pins:
                # if p != in_pin:
                pins_to_move.append(p)

        if out_wire:
            for p in pins_to_move:
                in_wire.disconnect_pin(p)
                out_wire.connect_pin(p)


def _bring_to_top(e, add_to_name, top_definition):
    """move the cable that is internal to the top level."""
    if "EDIF.identifier" in e:
        global mod_name_uid
        good = False
        if mod_name_uid == 45773:
            print(e["EDIF.identifier"])
            print(e)
            good = True
        if isinstance(e, Cable):
            e["EDIF.identifier"] = "cable_" + _get_unique_name_modifier()
        else:
            e["EDIF.identifier"] = "instance_" + _get_unique_name_modifier()

        if good:
            print(e["EDIF.identifier"])
    if isinstance(e, Cable):
        d = e.definition
        d.remove_cable(e)
    else:
        d = e.parent
        d.remove_child(e)
    # _rename_element(c)
    if(add_to_name != ''):
        e.name = add_to_name + "/" + e.name
    else:
        e.name = e.name
    if isinstance(e, Cable):
        top_definition.add_cable(e)
    else:
        top_definition.add_child(e)


# def simple_recursive_netlist_visualizer(netlist):
#     #TODO put this code somewhere where people can use it to visualize simple netlists
#     top_instance = netlist.top_instance
#     #should look something like this:
#     #top
#     #   child1
#     #       child1.child
#     #   child2
#     #       child2.child
#     def recurse(instance, depth):
#         s = depth * "\t"
#         print(s, instance.name, "(", instance.reference.name, ")")
#         for c in instance.reference.children:
#             recurse(c, depth + 1)

#     recurse(top_instance, 0)

def flatten(netlist):
    """
    starts at the top instance and brings all the different subelements to the top level.
    and port boundries are redone into one net.
    """

    # get all the sub instances of the top instance
    instance_queue = deque()
    name_queue = deque()
    top_instance = netlist.top_instance
    top_definition = top_instance.reference
    # put all of tops children on a stack
    for chld in top_definition.children:
        instance_queue.append(chld)
        name_queue.append('')

    to_remove = []
    # for each of the children on the stack
    while len(instance_queue) > 0:
        inst = instance_queue.popleft()
        parent_name = name_queue.popleft()
        # simple_recursive_netlist_visualizer(netlist)
        _bring_to_top(inst, parent_name, top_definition)
        if inst.reference.is_leaf():
            continue
        # put their children on the stack
        for child in inst.reference.children:
            instance_queue.append(child)
            name_queue.append(inst.name)
        temp_cables = []
        for cable in inst.reference.cables:
            temp_cables.append(cable)
        for cable in temp_cables:
            _bring_to_top(cable, inst.name, top_definition)
        for port in inst.reference.ports:
            _redo_connections(inst, port)
        to_remove.append(inst)
    for i in to_remove:
        top_definition.remove_child(i)


# mkdir export
# cd export
# write_edif
# write_xdc -constraints all leon3mp.xdc



import unittest

import spydrnet.parsers.verilog.verilog_tokens as vt
import spydrnet as sdn


class TestVerilogTokens(unittest.TestCase):


    def test_valid_identifier(self):
        identifiers = ["valid", "_valid", "\\^[0]should_be_valid ", "_1293i123"]
        for i in identifiers:
            assert vt.is_valid_identifier(i)

    def test_invalid_identifier(self):
        identifiers = ["not valid", "6invalid" "^[0]no_escape...", "\\space in escaped ", "", "!"]
        for i in identifiers:
            assert not vt.is_valid_identifier(i)

    def test_valid_number(self):
        numbers = ["123465789", "000000123", "0"]
        for n in numbers:
            assert vt.is_numeric(n)

    def test_invalid_number(self):
        #we could change this so that a decimal point is still a valid number if needed.
        numbers = ["123123.12903", "\\*#@(&$#@*$", "123 3901", ""]
        for n in numbers:
            assert not vt.is_numeric(n)

    def test_convert_string_to_direction(self):
        strings = ["input", "output", "inout", "unknown"]
        values = [sdn.Port.Direction.IN, sdn.Port.Direction.OUT, sdn.Port.Direction.INOUT, sdn.Port.Direction.UNDEFINED]
        for i in range(len(strings)):
            assert vt.string_to_port_direction(strings[i]) == values[i]